#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
from datetime import datetime, timedelta

# pip
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
import matplotlib.colors as mcolors
from matplotlib.colorbar import ColorbarBase

# teslakit
from ..plotting.custom_colors import GetClusterColors

# import constants
from ..plotting.config import _faspect, _fsize, _fdpi

# TODO: Plot_Complete, add xticks at last axs to show years

def axplot_bmus(ax, bmus, num_clusters, lab):
    'axes plot bmus series using colors'

    # colors
    wt_colors = GetClusterColors(num_clusters)

    # prepare nans as white 
    bmus = bmus.fillna(num_clusters+1)
    wt_colors = np.vstack([wt_colors, [0,0,0]])

    # bmus colors
    vp = wt_colors[bmus.astype(int)-1]

    # listed colormap
    ccmap = mcolors.ListedColormap(vp)

    # color series
    cbar_x = ColorbarBase(
        ax, cmap = ccmap, orientation='horizontal',
        norm = mcolors.Normalize(vmin=0, vmax=num_clusters),
    )
    cbar_x.set_ticks([])

    # customize axes
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylabel(lab, rotation=0, fontweight='bold', labelpad=35)

def axplot_series(ax, vv, ls, lc, lab, xticks=False):
    'axes plot variables series'

    # find start and end index
    nna = np.where(~np.isnan(vv))[0]
    x_0, x_1 = nna[0], nna[-1]

    # plot series
    ax.plot(nna, vv[nna], linestyle=ls, linewidth=0.5, color=lc)

    # customize axes
    ax.set_xlim(0, len(vv))

    # TODO add optional xticks option
    if not xticks:
        ax.set_xticks([])

    ax.yaxis.tick_right()
    ax.tick_params(axis='both', which='both', labelsize=7)
    ax.set_ylabel(lab, rotation=0, fontweight='bold', labelpad=35)

def Plot_Complete(xds, show=True):
    '''
    Plot complete data variables inside xds

    xds - xarray.Dataset: (time) AWT, MJO, DWT, Hs, Tp, Dir, SS, AT, MMSL, TWL
    '''
    # TODO: auto parameters inside xds

    # parameters
    n_cs_AWT = 6
    n_cs_MJO = 25
    n_cs_DWT = 42


    # plot figure
    fig, axs = plt.subplots(
        10, 1,
        figsize=(_faspect*_fsize, _fsize),
    )

    # AWT
    axplot_bmus(axs[0], xds.AWT, n_cs_AWT, 'AWT')

    # MJO
    axplot_bmus(axs[1], xds.MJO, n_cs_MJO, 'MJO')

    # DWT
    axplot_bmus(axs[2], xds.DWT, n_cs_DWT, 'DWT')

    # HS
    axplot_series(axs[3], xds.Hs, 'solid', 'cyan', 'Hs(m)')

    # TP
    axplot_series(axs[4], xds.Tp, 'solid', 'red', 'Tp(s)')

    # DIR
    axplot_series(axs[5], xds.Dir, 'dotted', 'black', 'Dir(º)')

    # SS
    if 'SS' in xds.variables:
        axplot_series(axs[6], xds.SS, 'solid', 'orange', 'SS(m)')
    else:
        axs[6].set_xticks([])
        axs[6].set_yticks([])
        axs[6].set_ylabel('SS(m)', rotation=0, fontweight='bold', labelpad=35)

    # AT
    axplot_series(axs[7], xds.AT, 'solid', 'purple', 'AT(m)')

    # MMSL
    axplot_series(axs[8], xds.MMSL, 'solid', 'green', 'MMSL(m)')

    # TWL
    axplot_series(axs[9], xds.TWL, 'solid', 'blue', 'TWL(m)')


    # suptitle
    fig.suptitle('Complete Data', fontweight='bold', fontsize=12)

    # reduce first 3 subplots
    gs = gridspec.GridSpec(10, 1, height_ratios = [1,1,1,3,3,3,3,3,3,3])
    for i in range(10):
        axs[i].set_position(gs[i].get_position(fig))

    # show and return figure
    if show: plt.show()
    return fig


def axplot_compare_histograms(ax, var_1, var_2, n_bins=40,
                      label_1='Historical', label_2='Simulation', ttl='',
                      alpha_1=0.9, alpha_2=0.7,
                      color_1='white', color_2='skyblue',
                      density=False):
    'axes plot histogram comparison between fit-sim variables'

    (_, bins, _) = ax.hist(var_1, n_bins, weights=np.ones(len(var_1)) / len(var_1),
            alpha=alpha_1, color=color_1, ec='k', label = label_1, density=density)

    ax.hist(var_2, bins=bins, weights=np.ones(len(var_2)) / len(var_2),
            alpha=alpha_2, color=color_2, ec='k', label = label_2, density=density)

    # customize axes
    ax.legend(prop={'size':8})
    ax.set_title(ttl)

    if density:
        ax.set_ylabel('Probability')

def Plot_FitSim_Histograms(data_fit, data_sim, vns, n_bins=40,
                           color_1='white', color_2='skyblue',
                           alpha_1=0.7, alpha_2=0.4,
                           label_1='Historical', label_2 = 'Simulation',
                           gs_1 = 1, gs_2 = None, supt=False, vns_lims={},
                           density=False, show=True):
    'Plots fit vs sim histograms for variables "vns"'

    # grid spec default number of columns
    if gs_2 == None: gs_2 = len(vns)

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize*gs_1/2.3))

    # grid spec
    gs = gridspec.GridSpec(gs_1, gs_2)  #, wspace=0.0, hspace=0.0)

    # variables
    cr, cc = 0, 0
    for c, vn in enumerate(vns):

        dh = data_fit[vn].values[:]; dh = dh[~np.isnan(dh)]
        ds = data_sim[vn].values[:]; ds = ds[~np.isnan(ds)]

        # variable max and min optional limits
        if vn in vns_lims.keys():
            vl_1, vl_2 = vns_lims[vn]
            dh = dh[np.where((dh >= vl_1) & (dh <= vl_2))[0]]
            ds = ds[np.where((ds >= vl_1) &( ds <= vl_2))[0]]

        ax = plt.subplot(gs[cr, cc])
        axplot_compare_histograms(
            ax, dh, ds, ttl=vn, density=density, n_bins=n_bins,
            color_1=color_1, color_2=color_2,
            alpha_1=alpha_1, alpha_2=alpha_2,
            label_1=label_1, label_2=label_2,
        )

        # grid spec counter
        cc+=1
        if cc >= gs_2:
            cc=0
            cr+=1

    # fig suptitle
    if supt:
        fig.suptitle(
            '{0} - {1} Comparison: {2}'.format(label_1, label_2, ', '.join(vns)),
            fontsize=13, fontweight = 'bold',
        )

    # show and return figure
    if show: plt.show()
    return fig

def Plot_LevelVariables_Histograms(data_hist, data_sim, label_1='Historical', label_2 = 'Simulation', show=True):
    'Plots histogram comparison (historical - simulation) for level related variables'

    # Compare histograms 
    Plot_FitSim_Histograms(
        data_hist, data_sim, ['level', 'AT', 'MMSL', 'TWL'],
        color_1='white', color_2='skyblue', alpha_1=0.9, alpha_2=0.7,
        label_1= label_1, label_2 = label_2,
        gs_1 = 2, gs_2 = 2,
        density=False, show=True
    )




from typing import List, Tuple, Union

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import BoundaryNorm, Colormap, ListedColormap


def get_list_of_colors_for_colormap(
    cmap: Union[str, Colormap], num_colors: int
) -> list:
    """
    Get a list of colors from a colormap.

    Parameters
    ----------
    cmap : Union[str, Colormap]
        The colormap to use.
    num_colors : int
        The number of colors to generate.

    Returns
    -------
    list
        A list of colors generated from the colormap.
    """

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)

    return [cmap(i) for i in range(0, 256, 256 // num_colors)]


def create_cmap_from_colors(
    color_list: List[str], name: str = "custom"
) -> colors.LinearSegmentedColormap:
    """
    Create a colormap from a list of hex colors.

    Parameters
    ----------
    color_list : List[str]
        List of hex color codes (e.g., ["#ff0000", "#00ff00"])
    name : str, optional
        Name for the colormap. Default is "custom".

    Returns
    -------
    colors.LinearSegmentedColormap
        A colormap created from the provided colors.
    """

    rgb_colors = [colors.hex2color(color) for color in color_list]

    return colors.LinearSegmentedColormap.from_list(name, rgb_colors, N=256)


def join_colormaps(
    cmap1: Union[str, List[str], Colormap],
    cmap2: Union[str, List[str], Colormap],
    name: str = "joined_cmap",
    range1: Tuple[float, float] = (0.0, 1.0),
    range2: Tuple[float, float] = (0.0, 1.0),
    value_range1: Tuple[float, float] = None,
    value_range2: Tuple[float, float] = None,
) -> Tuple[ListedColormap, BoundaryNorm]:
    """
    Join two colormaps into one, with value ranges specified for each.

    Parameters
    ----------
    cmap1, cmap2 : Union[str, List[str], Colormap]
        Input colormaps (name, list of hex codes, or Colormap object).
    name : str
        Name of the output colormap.
    range1, range2 : Tuple[float, float]
        Portion of each colormap to use (from 0 to 1).
    value_range1, value_range2 : Tuple[float, float]
        Value ranges in the data domain corresponding to each colormap.

    Returns
    -------
    ListedColormap
        Merged colormap object.
    BoundaryNorm
        Normalization for mapping data to colors.
    """

    # Convert cmap1 to a Colormap if needed
    if isinstance(cmap1, str):
        cmap1 = plt.get_cmap(cmap1)
    elif isinstance(cmap1, list):
        cmap1 = colors.LinearSegmentedColormap.from_list("cmap1", cmap1)

    if isinstance(cmap2, str):
        cmap2 = plt.get_cmap(cmap2)
    elif isinstance(cmap2, list):
        cmap2 = colors.LinearSegmentedColormap.from_list("cmap2", cmap2)

    # Get colors from each colormap
    colors1 = cmap1(np.linspace(range1[0], range1[1], 128))
    colors2 = cmap2(np.linspace(range2[0], range2[1], 128))
    newcolors = np.vstack((colors1, colors2))

    # Create corresponding boundaries in data space
    if value_range1 is not None and value_range2 is not None:
        bounds1 = np.linspace(value_range1[0], value_range1[1], 129)
        bounds2 = np.linspace(value_range2[0], value_range2[1], 129)
        all_bounds = np.sort(np.concatenate([bounds1[:-1], bounds2]))

        norm = BoundaryNorm(boundaries=all_bounds, ncolors=len(newcolors))

        return colors.ListedColormap(newcolors, name=name), norm
    else:
        return colors.ListedColormap(newcolors, name=name)


if __name__ == "__main__":
    # Join two named colormaps using only middle 80% of each
    cmap = join_colormaps("viridis", "plasma", range1=(0.1, 0.9), range2=(0.1, 0.9))

    # Join a named colormap with a list of colors
    cmap = join_colormaps("viridis", ["#ff0000", "#00ff00", "#0000ff"])

    # Join two lists of colors
    cmap = join_colormaps(["#ff0000", "#00ff00"], ["#0000ff", "#ffff00"])

    # Join with custom name and ranges
    cmap = join_colormaps(
        "viridis", "plasma", name="my_custom_cmap", range1=(0.0, 0.5), range2=(0.5, 1.0)
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import numpy as np
import matplotlib.pyplot as plt

# teslakit
from ..util.time_operations import date2datenum as d2d

# import constants
from .config import _faspect, _fsize, _fdpi

def axplot_histcompare(ax, var_fit, var_sim, color='green', n_bins=40,
                       label_1='Historical', label_2='Simulation'):
    'axes plot histogram comparison between fit-sim variables'

    (_, bins, _) = ax.hist(var_fit, n_bins, weights=np.ones(len(var_fit)) / len(var_fit),
            alpha=0.9, color='white', ec='k', label = label_1)

    ax.hist(var_sim, bins=bins, weights=np.ones(len(var_sim)) / len(var_sim),
            alpha=0.4, color=color, ec='k', label = label_2)

    # customize axes
    ax.legend(prop={'size':10})

def Plot_TideSeries(time, tide, ttl, ylabel, show=True):
    'Plots generic tide data time series'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))
    plt.plot(
        time, tide, '-k',
        linewidth = 0.04,
    )
    plt.xlim(time[0], time[-1])
    plt.title(ttl)
    plt.xlabel('time')
    plt.ylabel(ylabel)

    # show and return figure
    if show: plt.show()
    return fig

def Plot_WaterLevel(time, atide, show=True):
    'Plots water level temporal series'

    ttl = 'Meassured Water Level'
    ylab = 'water level (m)'

    return Plot_TideSeries(time, atide, ttl, ylab, show)

def Plot_AstronomicalTide(time, atide, show=True):
    'Plots astronomical tide temporal series'

    ttl = 'Astronomical Tide'
    ylab = 'tide (m)'

    return Plot_TideSeries(time, atide, ttl, ylab, show)

def Plot_ValidateTTIDE(time, atide, atide_ttide, show=True):
    'Compares astronomical tide and Utide prediction'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))
    plt.plot(
        time, atide, '-k',
        linewidth = 0.04,
        label = 'data'
    )
    plt.plot(
        time, atide_ttide, '-r',
        linewidth = 0.02,
        label = 'Utide model'
    )
    plt.xlim(time[0], time[-1])
    plt.title('Astronomical tide - UTIDE validation')
    plt.xlabel('time')
    plt.ylabel('tide (m)')
    axs.legend()

    # show and return figure
    if show: plt.show()
    return fig

def axplot_sea_level(ax, sea_level_time, sea_level_val, var_time, var_val,
                     var_name):
    'axes plot sea level and other variable: running mean, slr, etc'

    ax.plot(
        sea_level_time, sea_level_val, '-k',
        linewidth = 0.2, label = 'tide'
    )
    ax.plot(
        var_time, var_val, '-r',
        linewidth = 1, label = var_name
    )

    # customize axs
    ax.legend()
    ax.set_xlim(var_time[0], var_time[-1])
    ax.set_title('Tide - {0}'.format(var_name), fontweight='bold')
    ax.set_xlabel('time')
    ax.set_ylabel('Sea Level (mm)')

def Plot_Tide_SLR(time, tide, slr, show=True):
    'Plots gauge tide temporal series and SLR'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))

    axplot_sea_level(axs, time, tide, time, slr, 'Sea Level Rise')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Tide_RUNM(time, tide, rum, show=True):
    'Plots gauge tide temporal series and runm'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))

    axplot_sea_level(axs, time, tide, time, rum, 'Running Mean')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Tide_MMSL(tide_time, tide_val, mmsl_time, mmsl_val, show=True):
    'Plots gauge tide temporal series and mmsl'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))

    axplot_sea_level(axs, tide_time, tide_val, mmsl_time, mmsl_val, 'MMSL')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Validate_MMSL_tseries(
    mmsl_time, mmsl_data, mmsl_pred, mmsl_pred_quantiles=np.array([]), show=True):
    'Plots series comparison between mmsl data and mmsl predicted with nlm'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))
    plt.plot(
        mmsl_time, mmsl_data, '-k',
        linewidth = 0.7, label = 'Historical'
    )
    plt.plot(
        mmsl_time, mmsl_pred, '-r',
        linewidth = 0.7, label = 'Prediction'
    )
    axs.axhline(linestyle='--', linewidth=0.5, color='k')
    axs.legend()
    axs.set_xlim(mmsl_time[0], mmsl_time[-1])

    ttl = 'Monthly Mean Sea Level linear model'
    axs.set_title(ttl, fontweight='bold')
    axs.set_xlabel('time')
    axs.set_ylabel('MMSL (mm)')

    # optional: plot mmsl quantiles
    if mmsl_pred_quantiles.any():
        plt.fill_between(
            mmsl_time, mmsl_pred_quantiles[0,:], mmsl_pred_quantiles[1,:],
            color='grey', alpha=.5, label='Percentiles',
        )

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Validate_scatter(mmsl_data, mmsl_pred, xlabel='', ylabel='', show=True):
    'Plots scatter comparison between mmsl data and mmsl predicted with nlm'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize/1.5, _fsize/1.5))
    axs.plot(
        mmsl_data, mmsl_pred, 'or'
    )
    axs.axhline(linestyle='--', linewidth=0.5, color='k')
    axs.axvline(linestyle='--', linewidth=0.5, color='k')

    dmin = np.nanmin(mmsl_data)
    dmax = np.nanmax(mmsl_data)
    axs.plot([dmin, dmax], [dmin, dmax], '--k')

    axs.set_xlabel(xlabel)
    axs.set_ylabel(ylabel)
    axs.set_aspect('equal', 'box')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Validate_MMSL_scatter(mmsl_data, mmsl_pred, show=True):
    'Plots scatter comparison between mmsl data and mmsl predicted with nlm'

    return Plot_Validate_scatter(
        mmsl_data, mmsl_pred,
        xlabel = 'Historical MMSL (mm)',
        ylabel = 'Simulated MMSL (mm)',
        show = show,
    )

def Plot_MMSL_Prediction(mmsl_time, mmsl_data, show=True, label='Simulation'):
    'Plots predicted mmsl data'

    # parse time array to datetime
    mmsl_time = [d2d(x) for x in mmsl_time]

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))
    axs.plot(
        mmsl_time, mmsl_data, '-g',
        linewidth = 0.5, label = label,
    )
    axs.axhline(linestyle='--', linewidth=0.5, color='k')

    axs.legend()
    axs.set_xlim(mmsl_time[0], mmsl_time[-1])
    ttl = 'Monthly Mean Sea Level linear model prediction'
    axs.set_title(ttl, fontweight='bold')
    axs.set_xlabel('time')
    axs.set_ylabel('MMSL (mm)')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_MMSL_Histogram(mmsl_fit, mmsl_sim, show=True,
                        label_1='Historical', label_2='Simulation'):
    'Plots predicted mmsl data'

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize/2))
    axplot_histcompare(
        axs, mmsl_fit, mmsl_sim, label_1=label_1, label_2=label_2
    )

    axs.set_xlabel('MMSL (mm)')

    # show and return figure
    if show: plt.show()
    return fig




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import xarray as xr


def Calculate_MMSL(xda_tide, year_ini, year_end):
    '''
    Calculate monthly mean sea level

    xda_tide  - xarray.DataArray (time,)
    '''

    lout_mean = []
    lout_median = []
    lout_time = []
    for yy in range(year_ini, year_end+1):
        for mm in range(1,13):

            d1 = np.datetime64('{0:04d}-{1:02d}'.format(yy,mm))
            d2 = d1 + np.timedelta64(1, 'M')

            tide_sel_m = xda_tide.where(
                (xda_tide.time >= d1) & (xda_tide.time <= d2),
                drop = True)[:-2]
            time_sel = tide_sel_m.time.values

            if len(time_sel) >= 300:
                # mean, median and dates
                ts_mean = tide_sel_m.mean().values
                ts_median = tide_sel_m.median().values
                ts_time = time_sel[int(len(time_sel)/2)]

                lout_mean.append(ts_mean)
                lout_median.append(ts_median)
                lout_time.append(ts_time)


    # join output in xarray.Dataset
    xds_MMSL = xr.Dataset(
        {
            'mmsl':(('time',), lout_mean),
            'mmsl_median':(('time',), lout_median),
        },
        coords = {
            'time': lout_time
        }
    )

    return xds_MMSL




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import subprocess as sp

import numpy as np
import xarray as xr

# SWAN STAT LIBRARY
from .io import SwanIO_STAT, SwanIO_NONSTAT


# grid description template
d_grid_template = {
    'xpc': None,      # x origin
    'ypc': None,      # y origin
    'alpc': None,     # x-axis direction 
    'xlenc': None,    # grid length in x
    'ylenc': None,    # grid length in y
    'mxc': None,      # number mesh x
    'myc': None,      # number mesh y
    'dxinp': None,    # size mesh x
    'dyinp': None,    # size mesh y
}

# swan input parameters template
d_params_template = {
    'sea_level': None,
    'jonswap_gamma': None,
    'coords_spherical': None,  # None, 'GCM', 'CCM'
    'cdcap': None,
    'maxerr': None,            # None, 1, 2, 3
    'waves_period': None,      # 'PEAK', 'MEAN'
    'nested_bounds': None,     # 'CLOSED', 'OPEN'
}


class SwanMesh(object):
    'SWAN numerical model mesh'

    def __init__(self):

        self.depth = None   # bathymetry depth value (2D numpy.array)
        self.depth_fn = ''  # filename used in SWAN execution
        self.output_fn = ''  # output .mat file for mesh comp. grid

        # grid parameters
        self.cg = d_grid_template.copy()  # computational grid
        self.dg = d_grid_template.copy()  # depth grid
        self.dg_idla = 1  # 1/3 input swan parameter, handles read order at readinp

    def export_dat(self, p_case):
        'exports depth values to .dat file'

        p_export = op.join(p_case, self.depth_fn)
        np.savetxt(p_export, self.depth, fmt='%.2f')

    def get_XY(self):
        'returns mesh X, Y arrays from computational grid'

        # computational grid
        cg = self.cg

        x0 = cg['xpc']
        x1 = cg['xlenc'] + cg['xpc'] - cg['dxinp']
        xN = cg['mxc']
        X = np.linspace(x0, x1, xN)

        y0 = cg['ypc']
        y1 = cg['ylenc'] + cg['ypc'] - cg['dyinp']
        yN = cg['myc']
        Y = np.linspace(y0, y1, yN)

        return X, Y


class SwanProject(object):
    'SWAN numerical model project parameters, grids and information'

    def __init__(self, p_proj, n_proj):
        '''
        SWAN project information will be stored here

        http://swanmodel.sourceforge.net/online_doc/swanuse/node25.html
        '''

        self.p_main = op.join(p_proj, n_proj)    # project path
        self.name = n_proj                       # project name

        # sub folders 
        self.p_cases = op.join(self.p_main, 'cases')  # project cases

        # SWAN mesh: main 
        self.mesh_main = SwanMesh()
        self.mesh_main.depth_fn = 'depth_main.dat'
        self.mesh_main.output_fn = 'output_main.mat'

        # SWAN mesh: nest1
        self.mesh_nest1 = SwanMesh()
        self.mesh_nest1.depth_fn = 'depth_nest1.dat'
        self.mesh_nest1.output_fn = 'output_nest1.mat'
        self.run_nest1 = False

        # SWAN mesh: nest2
        self.mesh_nest2 = SwanMesh()
        self.mesh_nest2.depth_fn = 'depth_nest2.dat'
        self.mesh_nest2.output_fn = 'output_nest2.mat'
        self.run_nest2 = False

        # SWAN mesh: nest3
        self.mesh_nest3 = SwanMesh()
        self.mesh_nest3.depth_fn = 'depth_nest3.dat'
        self.mesh_nest3.output_fn = 'output_nest3.mat'
        self.run_nest3 = False

        # swan execution parameteres
        self.params = d_params_template.copy()

        # output points
        self.x_out = []
        self.y_out = []


class SwanWrap(object):
    'SWAN numerical model wrap for multi-case handling'

    def __init__(self, swan_proj, swan_io):
        '''
        swan_proj - SwanProject() instance, contains project parameters
        swan_io   - SwanIO_STAT / SwanIO_NONSTAT modules (auto from children)
        '''

        # set project and IO module
        self.proj = swan_proj           # swan project parameters
        self.io = swan_io(self.proj)     # swan input/output 

        # swan bin executable
        p_res = op.join(op.dirname(op.realpath(__file__)), 'resources')
        self.bin = op.abspath(op.join(p_res, 'swan_bin', 'swan_ser.exe'))

    def get_run_folders(self):
        'return sorted list of project cases folders'

        # TODO: will find previously generated cases... fix it 

        ldir = sorted(os.listdir(self.proj.p_cases))
        fp_ldir = [op.join(self.proj.p_cases, c) for c in ldir]

        return [p for p in fp_ldir if op.isdir(p)]

    def run_cases(self):
        'run all cases inside project "cases" folder'

        # TODO: improve log / check execution ending status

        # get sorted execution folders
        run_dirs = self.get_run_folders()

        for p_run in run_dirs:

            # run case main mesh
            self.run(p_run)

            # run case nested mesh (optional)
            r_ns = [
                self.proj.run_nest1,
                self.proj.run_nest2,
                self.proj.run_nest3,
            ]
            i_ns = [
                'input_nest1.swn',
                'input_nest2.swn',
                'input_nest3.swn'
            ]
            for r_n, i_n in zip(r_ns, i_ns):
                if r_n:
                    self.run(p_run, input_file=i_n)

            # log
            p = op.basename(p_run)
            print('SWAN CASE: {0} SOLVED'.format(p))

    def run(self, p_run, input_file='input.swn'):
        'Bash execution commands for launching SWAN'

        # aux. func. for launching bash command
        def bash_cmd(str_cmd, out_file=None, err_file=None):
            'Launch bash command using subprocess library'

            _stdout = None
            _stderr = None

            if out_file:
                _stdout = open(out_file, 'w')
            if err_file:
                _stderr = open(err_file, 'w')

            s = sp.Popen(str_cmd, shell=True, stdout=_stdout, stderr=_stderr)
            s.wait()

            if out_file:
                _stdout.flush()
                _stdout.close()
            if err_file:
                _stderr.flush()
                _stderr.close()

        # ln input file and run swan case
        cmd = 'cd {0} && ln -sf {1} INPUT && {2} INPUT'.format(
            p_run, input_file, self.bin)
        bash_cmd(cmd)

    def extract_output(self, mesh=None):
        '''
        exctract output from all cases

        return xarray.Dataset (uses new dim "case" to join output)
        '''

        # select main or nested mesh
        if mesh == None: mesh = self.proj.mesh_main

        # get sorted execution folders
        run_dirs = self.get_run_folders()

        # exctract output case by case and concat in list
        l_out = []
        for p_run in run_dirs:

            # read output file
            xds_case_out = self.io.output_case(p_run, mesh)
            l_out.append(xds_case_out)

        # concatenate xarray datasets (new dim: case)
        xds_out = xr.concat(l_out, dim='case')

        return(xds_out)

    def extract_output_points(self):
        '''
        exctract output from points all cases table_outpts.dat

        return xarray.Dataset (uses new dim "case" to join output)
        '''

        # get sorted execution folders
        run_dirs = self.get_run_folders()

        # exctract output case by case and concat in list
        l_out = []
        for p_run in run_dirs:

            # read output file
            xds_case_out = self.io.output_points(p_run)
            l_out.append(xds_case_out)

        # concatenate xarray datasets (new dim: case)
        xds_out = xr.concat(l_out, dim='case')

        return(xds_out)


class SwanWrap_STAT(SwanWrap):
    'SWAN numerical model wrap for STATIONARY multi-case handling'

    def __init__(self, swan_proj):
        super().__init__(swan_proj, SwanIO_STAT)

    def build_cases(self, waves_dataset):
        '''
        generates all files needed for swan stationary multi-case execution

        waves_dataset - pandas.dataframe with "n" boundary conditions setup
        [n x 4] (hs, per, dir, spr)
        '''

        # make main project directory
        self.io.make_project()

        # one stat case for each wave sea state
        for ix, (_, ws) in enumerate(waves_dataset.iterrows()):

            # build stat case 
            case_id = '{0:04d}'.format(ix)
            self.io.build_case(case_id, ws)


class SwanWrap_NONSTAT(SwanWrap):
    'SWAN numerical model wrap for NON STATIONARY multi-case handling'

    def __init__(self, swan_proj):
        super().__init__(swan_proj, SwanIO_NONSTAT)

    def build_cases(self, waves_event_list, storm_track_list=None,
                    make_waves=True, make_winds=True):
        '''
        generates all files needed for swan non-stationary multi-case execution

        waves_event_list - list waves events time series (pandas.DataFrame)
        also contains level, tide and wind (not storm track) variables
        [n x 8] (hs, per, dir, spr, U10, V10, level, tide)

        storm_track_list - list of storm tracks time series (pandas.DataFrame)
        storm_track generated winds have priority over waves_event winds
        [n x 6] (move, vf, lon, lat, pn, p0)
        '''

        # check user input: no storm tracks
        if storm_track_list == None:
            storm_track_list = [None] * len(waves_event_list)

        # make main project directory
        self.io.make_project()

        # one non-stationary case for each wave time series
        for ix, (wds, sds) in enumerate(
            zip(waves_event_list, storm_track_list)):

            # build stat case 
            case_id = '{0:04d}'.format(ix)
            self.io.build_case(
                case_id, wds, storm_track=sds,
                make_waves=make_waves, make_winds=make_winds
            )




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
from datetime import datetime

# pip
import numpy as np
import xarray as xr

def MJO_Categories(rmm1, rmm2, phase):
    '''
    Divides MJO data in 25 categories.

    rmm1, rmm2, phase - MJO parameters

    returns array with categories time series
    and corresponding rmm
    '''

    rmm = np.sqrt(rmm1**2 + rmm2**2)
    categ = np.empty(rmm.shape) * np.nan

    for i in range(1,9):
        s = np.squeeze(np.where(phase == i))
        rmm_p = rmm[s]

        # categories
        categ_p = np.empty(rmm_p.shape) * np.nan
        categ_p[rmm_p <=1] =  25
        categ_p[rmm_p > 1] =  i + 8*2
        categ_p[rmm_p > 1.5] =  i + 8
        categ_p[rmm_p > 2.5] =  i
        categ[s] = categ_p

    # get rmm_categ
    rmm_categ = {}
    for i in range(1,26):
        s = np.squeeze(np.where(categ == i))
        rmm_categ['cat_{0}'.format(i)] = np.column_stack((rmm1[s],rmm2[s]))

    return categ.astype(int), rmm_categ

def MJO_Phases(rmm1, rmm2):
    'calculates and returns MJO phases (1-8) and degrees'

    # mjo degrees phase
    degr = (np.arctan2(rmm2, rmm1))*360/(2*np.pi)
    degr[degr<0] = degr[degr<0] + 360

    # degree to mjo phase (1-8)
    phases_sector = [
        (5, [0, 45]),
        (6, [45, 90]),
        (7, [90, 135]),
        (8, [135, 180]),
        (1, [180, 225]),
        (2, [225, 270]),
        (3, [270, 315]),
        (4, [315, 360]),
    ]

    # calculate phase
    phase = np.zeros(len(degr))*np.nan
    for p, s in phases_sector:
        phase[np.where((degr[:]>s[0]) & (degr[:]<=s[1]))] = p
    phase = phase.astype(int)

    return phase, degr




import datetime
import os.path as op
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm

from ..core.constants import EARTH_RADIUS
from ..core.geo import geodesic_azimuth, geodesic_distance_azimuth, shoot
from ..datamining.mda import find_nearest_indices
from ..waves.superpoint import superpoint_calculation
from .tracks import (
    PATHS,
    get_vmean,
    historic_track_interpolation,
    historic_track_preprocessing,
    track_triming,
)

###############################################################################
# STOPMOTION functions
# functions to preprocess and interpolate coordinates at swan computational
# timesteps for any storm track according to the SHyTCWaves methodology units:
# 6-hour target segment preceded by 24-hour warmup segments

# storm2stopmotion --> parameterization of a storm track into segments
# stopmotion_interpolation --> generate stop-motion events
###############################################################################


def storm2stopmotion(df_storm: pd.DataFrame) -> pd.DataFrame:
    """
    Generate stopmotion segments from storm track.

    Parameters
    ----------
    df_storm : pd.DataFrame
        Storm track DataFrame containing:
        - lon : Longitude coordinates
        - lat : Latitude coordinates
        - p0 : Minimum central pressure (mbar)
        - vmax : Maximum sustained winds (kt)
        - rmw : Radius of maximum winds (nmile)
        - vmaxfill : Boolean indicating if winds were filled with estimates
        - rmwfill : Boolean indicating if RMW was filled with estimates

    Returns
    -------
    pd.DataFrame
        Stopmotion segments with columns:
        - vseg : Mean translational speed (kt)
        - dvseg : Speed variation (kt)
        - pseg : Segment pressure (mbar)
        - dpseg : Pressure variation (mbar)
        - wseg : Segment maximum winds (kt)
        - dwseg : Wind variation (kt)
        - rseg : Segment RMW (nmile)
        - drseg : RMW variation (nmile)
        - aseg : Azimuth from geographic North (degrees)
        - daseg : Azimuth variation (degrees)
        - lseg : Latitude (degrees)
        - laseg : Absolute latitude (degrees)
        - dlaseg : Latitude variation (degrees)
        - lon_w : Warmup origin longitude
        - lat_w : Warmup origin latitude
        - lon_i : Target origin longitude
        - lat_i : Target origin latitude
        - lon_t : Target endpoint longitude
        - lat_t : Target endpoint latitude

    Notes
    -----
    The function generates stopmotion segments methodology from 6-hour storm track segments:

    A. Warmup segment (24h):
        - 4 segments to define start/end coordinates
        - Defines {Vmean, relative angle}
        - Last 4th segment defines mean {Pmin, Wmax, Rmw}
        - Endpoint defines {lat}

    B. Target segment (6h):
        Defines variations {dP, dV, dW, dR, dAng}

    The absolute value of latitude is stored (start of target segment).
    Relative angle is referenced to geographic north (southern hemisphere
    is multiplied by -1).
    """

    # remove NaNs
    df_ = df_storm.dropna()
    df_["time"] = df_.index.values

    # constant segments variables
    lon = df_["lon"].values[:]
    lat = df_["lat"].values[:]
    pres = df_["p0"].values[:]  # [mbar]
    wind = df_["vmax"].values[:]  # [kt]
    rmw = df_["rmw"].values[:]  # [nmile]

    # generate stopmotion segments: 24h warmup + 6h target segments

    # timestep [hours]
    dt = np.diff(df_.index) / np.timedelta64(1, "h")

    # warmup 4-segments (24h) variables
    lo0 = np.full(df_.shape[0], np.nan)  # warmup x-coordinate
    la0 = np.full(df_.shape[0], np.nan)  # warmup y-coordinate
    vseg = np.full(df_.shape[0], np.nan)  # mean translational speed
    vxseg = np.full(df_.shape[0], np.nan)  # (dirx)
    vyseg = np.full(df_.shape[0], np.nan)  # (diry)
    aseg = np.full(df_.shape[0], np.nan)  # azimuth, geographic North

    # warmup last-segment (6h) variables
    pseg = np.full(df_.shape[0], np.nan)  # segment pressure
    wseg = np.full(df_.shape[0], np.nan)  # segment maxwinds
    rseg = np.full(df_.shape[0], np.nan)  # segment radii rmw
    lseg = np.full(df_.shape[0], np.nan)  # latitude (north hemisphere)
    laseg = np.full(df_.shape[0], np.nan)  # (absolute)

    # target segment (6h) variables
    lo1 = np.full(df_.shape[0], np.nan)  # target origin x-coordinate
    la1 = np.full(df_.shape[0], np.nan)  # idem y-coordinate
    dv = np.full(df_.shape[0], np.nan)  # translational speed variation
    dvx = np.full(df_.shape[0], np.nan)  # (dirx)
    dvy = np.full(df_.shape[0], np.nan)  # (diry)
    da = np.full(df_.shape[0], np.nan)  # azimuth variation
    dp = np.full(df_.shape[0], np.nan)  # pressure variation
    dw = np.full(df_.shape[0], np.nan)  # maxwinds variation
    dr = np.full(df_.shape[0], np.nan)  # radii rmw variation
    dl = np.full(df_.shape[0], np.nan)  # latitude variation
    dla = np.full(df_.shape[0], np.nan)  # (absolute)
    lo2 = np.full(df_.shape[0], np.nan)  # target endpoint x-coordinate
    la2 = np.full(df_.shape[0], np.nan)  # idem y-coordinate

    # loop
    for i in np.arange(1, dt.size):
        # get stopmotion endpoints coordinates (24h+6h)
        if i < 4:  # < four preceding segments
            # number of "missing" preceding segments to last 24h
            n_missing = 4 - i

            # last available preceding segment
            lon1, lon2 = lon[1], lon[0]
            lat1, lat2 = lat[1], lat[0]

            # distance of last available preceding segment
            arcl_h, gamma_h = geodesic_distance_azimuth(lat2, lon2, lat1, lon1)
            r = arcl_h * np.pi / 180.0 * EARTH_RADIUS  # distance [km]

            # shoot backwards to calculate (lo0,la0) of 24h preceding warmup
            dist = r * n_missing
            glon, glat, baz = shoot(lon2, lat2, gamma_h + 180, dist)

            # endpoint coordinates (-24h, 0h, 6h)
            lon_0, lon_i, lon_i1 = glon, lon[i], lon[i + 1]
            lat_0, lat_i, lat_i1 = glat, lat[i], lat[i + 1]

        if i >= 4:  # >= four preceding segments
            # endpoint coordinates (-24h, 0h, 6h)
            lon_0, lon_i, lon_i1 = lon[i - 4], lon[i], lon[i + 1]
            lat_0, lat_i, lat_i1 = lat[i - 4], lat[i], lat[i + 1]

        # segment endpoints
        lo0[i], lo1[i], lo2[i] = lon_0, lon_i, lon_i1
        la0[i], la1[i], la2[i] = lat_0, lat_i, lat_i1

        # warmup 4-segments (24h) variables
        _, vseg[i], vxseg[i], vyseg[i] = get_vmean(lat_0, lon_0, lat_i, lon_i, 24)
        aseg[i] = geodesic_azimuth(lat_0, lon_0, lat_i, lon_i)
        #        aseg[i] = calculate_azimut(lon_0, lat_0, lon_i, lat_i)

        # warmup last-segment (6h) variables
        pseg[i] = pres[i - 1]
        wseg[i] = wind[i - 1]
        rseg[i] = rmw[i - 1]
        lseg[i] = lat_i
        laseg[i] = np.abs(lat_i)

        # target segment (6h) variables
        _, v, vx, vy = get_vmean(lat_i, lon_i, lat_i1, lon_i1, dt[i : i + 1].sum())
        dv[i] = v - vseg[i]  # [km/h]
        dvx[i] = vx - vxseg[i]
        dvy[i] = vy - vyseg[i]
        dp[i] = pres[i] - pres[i - 1]  # [mbar]
        dw[i] = wind[i] - wind[i - 1]  # [kt]
        dr[i] = rmw[i] - rmw[i - 1]  # [nmile]
        dl[i] = lat_i1 - lat_i  # [º]
        dla[i] = np.abs(dl[i])

        # angle variation
        ang1 = aseg[i]
        ang2 = geodesic_azimuth(lat_i, lon_i, lat_i1, lon_i1)
        #        ang2 = calculate_azimut(lon_i, lat_i, lon_i1, lat_i1)
        dt_ang = ang2 - ang1  # [º]
        sign = np.sign(lseg[i])  # hemisphere: north (+), south (-)

        if (ang2 > ang1) & (dt_ang < 180):
            da[i] = sign * (dt_ang)
        elif (ang2 > ang1) & (dt_ang > 180):
            da[i] = sign * (dt_ang - 360)
        elif (ang2 < ang1) & (dt_ang > -180):
            da[i] = sign * (dt_ang)
        elif (ang2 < ang1) & (dt_ang < -180):
            da[i] = sign * (dt_ang + 360)

    # add to dataframe
    df_["vseg"] = vseg / 1.852  # [kt]
    #    df_['vxseg'] = vxseg / 1.852
    #    df_['vyseg'] = vyseg / 1.852
    df_["dvseg"] = dv / 1.852
    #    df_['dvxseg'] = dvx / 1.852
    #    df_['dvyseg'] = dvy / 1.852
    df_["pseg"] = pseg  # [mbar]
    df_["dpseg"] = dp
    df_["wseg"] = wseg  # [kt, 1-min avg]
    df_["dwseg"] = dw
    df_["rseg"] = rseg  # [nmile]
    df_["drseg"] = dr
    df_["aseg"] = aseg  # [º]
    df_["daseg"] = da
    df_["lseg"] = lseg  # [º]
    df_["laseg"] = laseg
    df_["dlaseg"] = dla
    df_["lon_w"] = lo0  # warmup origin
    df_["lat_w"] = la0
    df_["lon_i"] = lo1  # target origin
    df_["lat_i"] = la1
    df_["lon_t"] = lo2  # target endpoint
    df_["lat_t"] = la2

    return df_


def stopmotion_interpolation(
    df_seg: pd.DataFrame,
    st: pd.DataFrame = None,
    t_warm: int = 24,
    t_seg: int = 6,
    t_prop: int = 42,
) -> Tuple[List[pd.DataFrame], List[pd.DataFrame]]:
    """
    Generate SWAN cases in cartesian coordinates from stopmotion parameterized segments.

    Parameters
    ----------
    df_seg : pd.DataFrame
        Stopmotion segments containing:
        - vseg : Mean translational speed (kt)
        - pseg : Segment pressure (mbar)
        - wseg : Maximum winds (kt)
        - rseg : RMW (nmile)
        - laseg : Absolute latitude (degrees)
        - dvseg : Speed variation (kt)
        - dpseg : Pressure variation (mbar)
        - dwseg : Wind variation (kt)
        - drseg : RMW variation (nmile)
        - daseg : Azimuth variation (degrees)

    st : pd.DataFrame, optional
        Real storm track data. If None, MDA segments are used (unrelated to historic tracks).
    t_warm : int, optional
        Warmup period in hours. Default is 24.
    t_seg : int, optional
        Target period in hours. Default is 6.
    t_prop : int, optional
        Propagation period in hours. Default is 42.

    Returns
    -------
    Tuple[List[pd.DataFrame], List[pd.DataFrame]]
        Two lists containing:

        1. List of storm track DataFrames with columns:
           - x, y : Cartesian coordinates (m)
           - lon, lat : Geographic coordinates (degrees)
           - vf : Translation speed (kt)
           - vfx, vfy : Velocity components (kt)
           - pn : Surface pressure (1013 mbar)
           - p0 : Minimum central pressure (mbar)
           - vmax : Maximum winds (kt)
           - rmw : RMW (nmile)
           - latitude : Latitude with hemisphere sign

        2. List of empty wave event DataFrames with columns:
           - hs, t02, dir, spr : Wave parameters
           - U10, V10 : Wind components
           - level, tide : Water level parameters

    Notes
    -----
    The function generates SWAN cases in cartesian coordinates following SHyTCWaves configuration:

    A. Warmup period (24h):
       Over the negative x-axis ending at (x,y)=(0,0)

    B. Target period (6h):
       Starting at (x,y)=(0,0)

    C. Propagation period (42h):
       No track coordinates (no wind forcing)
    """

    # sign hemisphere (+north, -south)
    if isinstance(st, pd.DataFrame):  # historic track
        sign = np.sign(st["lat"][0])
        method = st.attrs["method"]
        center = st.attrs["center"]
    else:  # mda cases
        sign = 1
        method = "mda"
        center = "mda"

    # remove NaNs
    df = df_seg.dropna()

    # number of stopmotion events
    N = df.shape[0]

    # list of SWAN cases (paramterized events)
    st_list, we_list = [], []
    for i in range(N):
        seg_i = df.iloc[i]

        # stopmotion unit parameters
        vseg = seg_i["vseg"] * 1.852  # [km/h]
        dvseg = seg_i["dvseg"] * 1.852
        pseg = seg_i["pseg"]  # [mbar]
        dpseg = seg_i["dpseg"]
        wseg = seg_i["wseg"]  # [kt, 1-min avg]
        dwseg = seg_i["dwseg"]
        rseg = seg_i["rseg"]  # [nmile]
        drseg = seg_i["drseg"]
        daseg = seg_i["daseg"]  # [º]
        laseg = seg_i["laseg"]  # [º]

        # vmean criteria for SWAN computational timestep [minutes]
        seg_vmean = vseg + dvseg
        if (vseg > 20) or (seg_vmean > 20):
            dt_comp = 10
        else:
            dt_comp = 20

        # time array for SWAN input
        ts = t_warm + t_seg + t_prop  # [h] simulation duration
        ts = np.asarray(ts) * 60 / dt_comp  # [] intervals of computation

        ts_warmup = int(t_warm * 60 / dt_comp)
        ts_segment = int(t_seg * 60 / dt_comp)

        # random initial date
        date_ini = pd.Timestamp(1999, 12, 31, 0)
        time_input = pd.date_range(
            date_ini, periods=int(ts), freq="{0}MIN".format(dt_comp)
        )
        time_input = np.array(time_input)

        # vortex input variables
        x = np.full(int(ts), np.nan)  # [m]
        y = np.full(int(ts), np.nan)  # [m]
        vmean = np.full(int(ts), np.nan)  # [km/h]
        ut = np.full(int(ts), np.nan)
        vt = np.full(int(ts), np.nan)
        p0 = np.full(int(ts), np.nan)  # [mbar]
        vmax = np.full(int(ts), np.nan)  # [kt, 1-min avg]
        rmw = np.full(int(ts), np.nan)  # [nmile]
        lat = np.full(int(ts), np.nan)  # [º]

        # (A) preceding 24h segment: over negative x-axis ending at (x,y)=(0,0)

        for j in np.arange(0, ts_warmup):
            if j == 0:
                x[j] = -vseg * 24 * 10**3
            else:
                x[j] = x[j - 1] + vseg * (dt_comp / 60) * 10**3
            y[j] = 0
            vmean[j] = vseg
            ut[j] = vseg
            vt[j] = 0
            p0[j] = pseg
            vmax[j] = wseg
            rmw[j] = rseg
            lat[j] = laseg

        # (B) target 6h segment: starting at (x,y)=(0,0)

        for j in np.arange(ts_warmup, ts_warmup + ts_segment):
            vel = vseg + dvseg  # [km/h]
            velx = vel * np.sin((daseg * sign + 90) * np.pi / 180)
            vely = vel * np.cos((daseg * sign + 90) * np.pi / 180)

            x[j] = x[j - 1] + velx * (dt_comp / 60) * 10**3
            y[j] = y[j - 1] + vely * (dt_comp / 60) * 10**3
            vmean[j] = vel
            ut[j] = velx
            vt[j] = vely
            p0[j] = pseg + dpseg
            vmax[j] = wseg + dwseg
            rmw[j] = rseg + drseg
            lat[j] = laseg

        # (C) propagation 42h segment: remaining values of data arrays

        # store dataframe
        st_seg = pd.DataFrame(
            index=time_input,
            columns=["x", "y", "vf", "vfx", "vfy", "pn", "p0", "vmax", "rmw", "lat"],
        )

        st_seg["x"] = x  # [m]
        st_seg["y"] = y
        st_seg["lon"] = x  # (idem for plots)
        st_seg["lat"] = y
        st_seg["vf"] = vmean / 1.852  # [kt]
        st_seg["vfx"] = ut / 1.852
        st_seg["vfy"] = vt / 1.852
        st_seg["pn"] = 1013  # [mbar]
        st_seg["p0"] = p0
        st_seg["vmax"] = vmax  # [kt]
        st_seg["rmw"] = rmw  # [nmile]
        st_seg["latitude"] = lat * sign  # [º]

        # add metadata
        st_seg.attrs = {
            "method": method,
            "center": center,
            "override_dtcomp": "{0} MIN".format(dt_comp),
            "x0": 0,
            "y0": 0,
            "p0": "mbar",
            "vf": "kt",
            "vmax": "kt, 1-min avg",
            "rmw": "nmile",
        }

        # append to stopmotion event list
        st_list.append(st_seg)

        # generate wave event (empty)
        we = pd.DataFrame(
            index=time_input, columns=["hs", "t02", "dir", "spr", "U10", "V10"]
        )
        we["level"] = 0
        we["tide"] = 0
        we_list.append(we)

    return st_list, we_list


###############################################################################
# STOPMOTION ensemble
# functions that collect 6h-segments from library cases to obtain the hybrid
# storm track (analogue or closest to the real track); those segments are
# rotated and assigned time-geographical coordinates. The ensemble and the
# envelope is calculate at each control point
###############################################################################


def find_analogue(
    df_library: pd.DataFrame, df_case: pd.DataFrame, ix_weights: List[float]
) -> np.ndarray:
    """
    Find the minimum distance in a 10-dimensional normalized space.

    Parameters
    ----------
    df_library : pd.DataFrame
        Library parameters DataFrame containing the 10 SHyTCWaves parameters:
        {pseg, vseg, wseg, rseg, dp, dv, dw, dr, dA, lat}
    df_case : pd.DataFrame
        Target case parameters DataFrame with same structure as df_library.
    ix_weights : List[float]
        Weight factors for each parameter dimension.

    Returns
    -------
    np.ndarray
        Indices of nearest points in the library for each case point.

    Notes
    -----
    The function finds the minimum distance in a normalized space corresponding
    to the SHyTCWaves 10 parameters:
    - pseg : Segment pressure
    - vseg : Mean translational speed
    - wseg : Maximum winds
    - rseg : RMW
    - dp : Pressure variation
    - dv : Speed variation
    - dw : Wind variation
    - dr : RMW variation
    - dA : Azimuth variation
    - lat : Latitude
    """

    # remove NaNs from storm segments
    df_case = df_case.dropna()
    data_case = df_case[
        [
            "daseg",
            "dpseg",
            "pseg",
            "dwseg",
            "wseg",
            "dvseg",
            "vseg",
            "drseg",
            "rseg",
            "laseg",
        ]
    ].values

    # library segments parameter
    data_lib = df_library[
        [
            "daseg",
            "dpseg",
            "pseg",
            "dwseg",
            "wseg",
            "dvseg",
            "vseg",
            "drseg",
            "rseg",
            "laseg",
        ]
    ].values
    ix_directional = [0]

    # get indices of nearest n-dimensional point
    ix_near = find_nearest_indices(
        query_points=data_case,
        reference_points=data_lib,
        directional_indices=ix_directional,
        weights=ix_weights,
    )

    return ix_near


def analogue_endpoints(df_seg: pd.DataFrame, df_analogue: pd.DataFrame) -> pd.DataFrame:
    """
    Add segment endpoint coordinates by looking up the real target segment.

    Parameters
    ----------
    df_seg : pd.DataFrame
        Parameterized historical storm track DataFrame containing:
        - lon, lat : Target origin coordinates
        - aseg : Warmup azimuth
        - daseg : Target azimuth variation
    df_analogue : pd.DataFrame
        Analogue segments from library containing:
        - vseg : Mean translational speed (kt)
        - dvseg : Speed variation (kt)
        - daseg : Target azimuth variation

    Returns
    -------
    pd.DataFrame
        Updated analogue segments DataFrame with added columns:
        - aseg : Warmup azimuth
        - lon_w, lat_w : Warmup origin coordinates
        - lon_i, lat_i : Target origin coordinates (from df_seg)
        - lon_t, lat_t : Target endpoint coordinates

    Notes
    -----
    The function:
    1. Calculates warmup origin by shooting backwards 24h from target origin
    2. Uses target origin from historical track
    3. Calculates target endpoint by shooting forwards 6h from target origin
    4. Accounts for hemisphere when calculating angles
    5. Converts longitudes to [0-360°] convention
    """

    # remove NaNs
    df_seg = df_seg[~df_seg.isna().any(axis=1)]
    # df_analogue = df_analogue[~df_analogue.isna().any(axis=1)]

    # get hemisphere
    # relative angles are multiplied by "sign" to account for hemisphere
    sign = np.sign(df_seg.lat.values[0])

    # get historic variables
    lon0 = df_seg.lon.values  # target origin coords
    lat0 = df_seg.lat.values
    aseg = df_seg.aseg.values  # warmup azimuth
    daseg = df_seg.daseg.values * sign  # target azimuth variation

    # get analogue variables
    vseg1_an = df_analogue.vseg.values  # warmup velocity [kt]
    dvseg_an = df_analogue.dvseg.values
    vseg2_an = vseg1_an + dvseg_an  # target velocity
    daseg_an = df_analogue.daseg.values * sign  # target azimuth variation

    # new variables
    az1_an = np.full(lon0.shape, np.nan)
    glon1_an = np.full(lon0.shape, np.nan)
    glon2_an = np.full(lon0.shape, np.nan)
    glat1_an = np.full(lon0.shape, np.nan)
    glat2_an = np.full(lon0.shape, np.nan)

    for i in range(df_seg.shape[0]):
        # azimuth angles for ensemble
        az2 = aseg[i] + daseg[i]  # target azimuth (historic-fixed)
        az1 = az2 - daseg_an[i]  # warmup azimuth (stopmotion analogue)

        # shoot backwards to warmup origin
        dist1 = vseg1_an[i] * 1.852 * 24  # [km]
        glon1, glat1, baz = shoot(lon0[i], lat0[i], az1 + 180, dist1)

        # shoot forwards to target endpoint
        dist2 = vseg2_an[i] * 1.852 * 6  # [km]
        glon2, glat2, baz = shoot(lon0[i], lat0[i], az2 + 180 - 180, dist2)

        # store
        az1_an[i] = az1
        glon1_an[i] = glon1
        glon2_an[i] = glon2
        glat1_an[i] = glat1
        glat2_an[i] = glat2

    # longitude convention
    glon1_an[glon1_an < 0] += 360
    glon2_an[glon2_an < 0] += 360

    # add to dataframe
    df_analogue["aseg"] = az1_an
    df_analogue["lon_w"] = glon1_an  # warmup origin
    df_analogue["lat_w"] = glat1_an
    df_analogue["lon_i"] = df_seg.lon.values
    df_analogue["lat_i"] = df_seg.lat.values
    df_analogue["lon_t"] = glon2_an
    df_analogue["lat_t"] = glat2_an

    return df_analogue


def stopmotion_st_bmu(
    df_analogue: pd.DataFrame,
    df_seg: pd.DataFrame,
    st: pd.DataFrame,
    cp_lon_ls: List[float],
    cp_lat_ls: List[float],
    max_dist: float = 60,
    list_out: bool = False,
    tqdm_out: bool = False,
    text_out: bool = True,
    mode: str = "",
) -> xr.Dataset:
    """
    Extract bulk wave parameters from library analogue cases.

    Parameters
    ----------
    df_analogue : pd.DataFrame
        Analogue prerun segments from library.
    df_seg : pd.DataFrame
        Storm 6h-segments parameters.
    st : pd.DataFrame
        Storm track interpolated every 6h.
    cp_lon_ls : List[float]
        Control point longitude coordinates.
    cp_lat_ls : List[float]
        Control point latitude coordinates.
    max_dist : float, optional
        Maximum distance (km) to extract closest node. Default is 60.
    list_out : bool, optional
        Whether to return list of datasets instead of merged dataset. Default is False.
    tqdm_out : bool, optional
        Whether to show progress bar. Default is False.
    text_out : bool, optional
        Whether to show text output. Default is True.
    mode : str, optional
        High or low resolution library indices. Default is "".

    Returns
    -------
    xr.Dataset
        Wave directional spectra with dimensions:
        - case : Storm segments
        - point : Control points
        - time : Time steps

        Contains variables:
        - hs : Significant wave height
        - tp : Peak period
        - lon, lat : Control point coordinates
        - ix_near : Nearest point indices
        - pos_nonan : Valid point mask
        - bmu : Best matching unit indices
        - hsbmu : Maximum Hs per point and time
        - tpbmu : Tp at maximum Hs
        - hswath : Maximum Hs per point
        - tswath : Tp at maximum Hs per point

    Notes
    -----
    The function:
    1. Accesses library analogue cases for a given storm track
    2. Calculates distance and angle from target segment origin to control points
    3. Extracts wave parameters at closest nodes for each analogue segment
    4. Computes bulk parameter envelope and swath
    5. Handles both high and low resolution libraries
    """

    # remove NaN
    df_seg = df_seg[~df_seg.isna().any(axis=1)]
    df_analogue = df_analogue[~df_analogue.isna().any(axis=1)]

    # assign time
    df_seg["time"] = df_seg.index.values

    # get hemisphere
    sign = np.sign(df_seg.lat.values[0])

    xds_list = []

    if tqdm_out:
        array = tqdm(range(df_seg.shape[0]))
    else:
        array = range(df_seg.shape[0])

    for iseg in array:  # each segment
        # get storm segment 'i'
        df_icase = df_seg.iloc[iseg]
        df_ianalogue = df_analogue.iloc[iseg]
        iseg_analogue = df_ianalogue.name  # analogue id
        aseg = df_ianalogue.aseg  # analogue warmseg azimuth (N)

        # ---------------------------------------------------------------------
        # get storm coordinates at "seg_time"
        seg_time = np.datetime64(df_icase.time)  # timestamp to datetime64
        ind_time_st = np.where(st.index.values == seg_time)[0][0]

        # storm coordinates (real storm eye)
        st_lon = st.iloc[ind_time_st].lon  # target origin longitude
        st_lat = st.iloc[ind_time_st].lat  # target origin latitude
        st_time = st.index.values[ind_time_st]  # time coordinates

        # ---------------------------------------------------------------------
        # get cp coordinates in relative system (radii, angle)
        cp_dist_ls, cp_ang_ls = get_cp_radii_angle(
            st_lat, st_lon, cp_lat_ls, cp_lon_ls, sign, aseg
        )

        # get SWAN output mask indices (rad, ang)
        mask_rad, mask_ang = get_mask_radii_angle(iseg_analogue, mode=mode)

        # find closest index
        ix_near = find_nearest(cp_dist_ls, cp_ang_ls, mask_rad, mask_ang)
        pos_nonan = np.abs(mask_rad[ix_near] - cp_dist_ls) <= max_dist

        # ---------------------------------------------------------------------
        # load library Hs reconstructed
        xds_rec = xr.open_dataset(PATHS["SHYTCWAVES_BULK"])
        xds_rec["time"] = pd.date_range("2000-01-01", periods=48, freq="1H")

        # extract HS,TP at closest points (case,point,time)
        hs_arr = np.full((ix_near.size, xds_rec.time.values.size), np.nan)
        hs_arr[pos_nonan, :] = (
            xds_rec.hs.isel(case=iseg_analogue, point=ix_near[pos_nonan]).load().values
        )

        tp_arr = np.full((ix_near.size, xds_rec.time.values.size), np.nan)
        tp_arr[pos_nonan, :] = (
            xds_rec.tp.isel(case=iseg_analogue, point=ix_near[pos_nonan]).load().values
        )

        # time array
        hour_intervals = xds_rec.time.size
        time = [st_time + np.timedelta64(1, "h") * i for i in range(hour_intervals)]
        time_array = np.array(time)

        # store dataset
        xds = xr.Dataset(
            {
                "hs": (("case", "point", "time"), np.expand_dims(hs_arr, axis=0)),
                "tp": (("case", "point", "time"), np.expand_dims(tp_arr, axis=0)),
                "lon": (("point"), cp_lon_ls),
                "lat": (("point"), cp_lat_ls),
                "ix_near": (("case", "point"), np.expand_dims(ix_near, axis=0)),
                "pos_nonan": (("case", "point"), np.expand_dims(pos_nonan, axis=0)),
            },
            coords={
                "case": [iseg],
                "time": time_array,
            },
        )
        xds_list.append(xds)

    if list_out:
        xds_out = xds_list
    else:
        # merge
        xds_out = xr.merge(xds_list)
        if text_out:
            print("Merging bulk envelope...", datetime.datetime.now())

        # add envelope variables
        xds_bmu = xds_out.copy()
        hsval = xds_bmu.hs.values[:]
        hsval[np.isnan(hsval)] = 0  # remove nan
        bmu = np.argmax(hsval, axis=0).astype(float)  # bmu indices
        hsmax = np.sort(hsval, axis=0)[-1, :, :]  # max over 'case'

        # bmu, hs
        bmu[hsmax == 0] = np.nan  # restitute nans
        hsmax[hsmax == 0] = np.nan

        xds_out["bmu"] = (("point", "time"), bmu)
        xds_out["hsbmu"] = (("point", "time"), hsmax)

        # tp
        tpmax = np.full(xds_out.hsbmu.shape, np.nan)
        nanmask = ~np.isnan(bmu)
        mesht, meshp = np.meshgrid(
            np.arange(0, xds_out.time.size), np.arange(0, xds_out.point.size)
        )

        tpmax[nanmask] = xds_out.tp.values[
            bmu.ravel()[nanmask.ravel()].astype("int64"),
            meshp.ravel()[nanmask.ravel()],
            mesht.ravel()[nanmask.ravel()],
        ]
        xds_out["tpbmu"] = (("point", "time"), tpmax)

        # add swath variables
        hh = hsmax.copy()
        hh[np.isnan(hh)] = 0
        posw = np.argmax(hh, axis=1)

        xds_out["hswath"] = (("point"), hsmax[np.arange(0, xds_out.point.size), posw])
        xds_out["tswath"] = (("point"), tpmax[np.arange(0, xds_out.point.size), posw])

    return xds_out


def stopmotion_st_spectra(
    df_analogue: pd.DataFrame,
    df_seg: pd.DataFrame,
    st: pd.DataFrame,
    cp_lon_ls: List[float],
    cp_lat_ls: List[float],
    cp_names: List[str] = [],
    max_dist: float = 60,
    list_out: bool = False,
    tqdm_out: bool = False,
    text_out: bool = True,
    mode: str = "",
) -> Tuple[xr.Dataset, xr.Dataset]:
    """
    Function to access the library analogue cases for a given storm track,
    calculate distance and angle from the target segment origin to the control
    point (relative coordinate system), and extract the directional wave
    spectra at the closest node (for every analogue segment)

    Parameters
    ----------
    df_analogue : pd.DataFrame
        Analogue prerun segments from library.
    df_seg : pd.DataFrame
        Storm 6h-segments parameters.
    st : pd.DataFrame
        Storm track interpolated every 6h.
    cp_lon_ls : List[float]
        Control point geographical coordinates.
    cp_lat_ls : List[float]
        Control point geographical coordinates.
    cp_names : List[str], optional
        Control point names. Default is [].
    max_dist : float, optional
        Maximum distance [km] to extract closest node. Default is 60.
    list_out : bool, optional
        Whether to list output. Default is False.
    tqdm_out : bool, optional
        Whether to use tqdm. Default is False.
    text_out : bool, optional
        Whether to print text. Default is True.
    mode : str, optional
        Mode. Default is "".

    Returns
    -------
    Tuple[xr.Dataset, xr.Dataset]
        - xds_spec : Wave directional spectra (dim 'case')
        - xds_bmu : BMU indices

    Notes
    -----
    The function:
    1. Removes NaN values from df_seg and df_analogue
    2. Assigns time to df_seg
    3. Gets hemisphere sign
    4. Gets bmu (wavespectra reconstructed)
    5. Opens seg_sim dataset
    """

    # remove NaN
    df_seg = df_seg[~df_seg.isna().any(axis=1)]
    df_analogue = df_analogue[~df_analogue.isna().any(axis=1)]

    # assign time
    df_seg["time"] = df_seg.index.values

    # get hemisphere
    sign = np.sign(df_seg.lat.values[0])

    # get bmu (wavespectra reconstructed)
    # it provides 'time,bmu,ix_near,pos_nonan' (point,time)
    xds_bmu = stopmotion_st_bmu(
        df_analogue=df_analogue,
        df_seg=df_seg,
        st=st,
        cp_lon_ls=cp_lon_ls,
        cp_lat_ls=cp_lat_ls,
        max_dist=max_dist,
        list_out=list_out,
        tqdm_out=tqdm_out,
        text_out=text_out,
        mode=mode,
    )

    # spectral energy
    seg_sim = xr.open_dataset(
        op.join(PATHS["SHYTCWAVES_SPECTRA"], "0000/spec_outpts_main.nc")
    )
    efth_arr = np.full(
        (
            seg_sim.frequency.size,
            seg_sim.direction.size,
            xds_bmu.point.size,
            xds_bmu.time.size,
        ),
        np.nan,
    )  # 38,72,cp,t

    if tqdm_out:
        array = tqdm(range(xds_bmu.case.size))
    else:
        array = range(xds_bmu.case.size)
    for iseg in array:
        # get storm segment 'i'
        df_icase = df_seg.iloc[iseg]
        df_ianalogue = df_analogue.iloc[iseg]
        iseg_analogue = df_ianalogue.name  # analogue id

        # ---------------------------------------------------------------------
        # get storm coordinates at "seg_time"
        seg_time = np.datetime64(df_icase.time)  # timestamp to datetime64
        st_time = st.index.values[st.index.values == seg_time][0]

        # ---------------------------------------------------------------------
        # get analogue segment from library
        filename = "spec_outpts_main.nc"
        p_analogue = op.join(
            PATHS["SHYTCWAVES_SPECTRA"], f"{iseg_analogue:04d}", filename
        )
        # load file
        seg_sim = xr.open_dataset(p_analogue)  # freq,dir,2088,48

        # time array
        hour_intervals = seg_sim.time.size
        time = [st_time + np.timedelta64(1, "h") * i for i in range(hour_intervals)]
        time_array = np.array(time)

        # get intersect time iseg vs xds_bmu
        _, ix_time_st, ix_time_shy = np.intersect1d(
            time_array, xds_bmu.time.values, return_indices=True
        )

        # find all closest grid points
        shy_inear = xds_bmu.ix_near.values[iseg, :].astype("int64")  # case,point

        # find bmu indices for iseg
        in_pt, in_t = np.where(xds_bmu.bmu.values == iseg)

        # get indices of casei
        in_t_ = in_t - ix_time_shy[0]

        # reorder spectral directions
        base = 5
        if mode == "_lowres":
            base = 10  # depends on the library dirs delta
        efth_case = seg_sim.isel(point=shy_inear)  # .isel(point=in_pt, time=in_t_)
        if sign < 0:
            efth_case["direction"] = 360 - seg_sim.direction.values
            new_dirs = np.round(
                efth_case.direction.values + base * round(df_icase.aseg / base) + 90, 1
            )
        else:
            new_dirs = np.round(
                efth_case.direction.values + base * round(df_icase.aseg / base) - 90, 1
            )
        new_dirs = np.mod(new_dirs, 360)
        new_dirs[new_dirs > 270] -= 360
        efth_case["direction"] = new_dirs
        efth_case = efth_case.sel(direction=seg_sim.direction.values)

        # insert spectral values for bmu=iseg
        efth_arr[:, :, in_pt, in_t] = efth_case.efth.values[:, :, in_pt, in_t_]

    if text_out:
        print("Inserting envelope spectra...", datetime.datetime.now())

    # store dataset
    xds_spec = xr.Dataset(
        {
            "efth": (("freq", "dir", "point", "time"), efth_arr),
            "lon": (("point"), np.array(cp_lon_ls)),
            "lat": (("point"), np.array(cp_lat_ls)),
            "station": (("point"), np.array(cp_names)),
        },
        coords={
            "freq": seg_sim.frequency.values,
            "dir": seg_sim.direction.values,
            "point": xds_bmu.point.values,
            "time": xds_bmu.time.values,
        },
    )

    return xds_spec, xds_bmu


def get_cp_radii_angle(
    st_lat: float,
    st_lon: float,
    cp_lat_ls: List[float],
    cp_lon_ls: List[float],
    sign: int,
    aseg: float,
) -> Tuple[List[float], List[float]]:
    """
    Extract control point distances and angles in the relative coordinate system.

    Parameters
    ----------
    st_lat : float
        Storm center latitude.
    st_lon : float
        Storm center longitude.
    cp_lat_ls : List[float]
        Control point latitudes.
    cp_lon_ls : List[float]
        Control point longitudes.
    sign : int
        Hemisphere indicator: 1 for north, -1 for south.
    aseg : float
        Azimuth of the analogue warm segment (from geographic north).

    Returns
    -------
    Tuple[List[float], List[float]]
        Two lists containing:
        - cp_dist_ls : Distances from storm center to control points (km)
        - cp_ang_ls : Angles from storm center to control points (degrees)

    Notes
    -----
    The function:
    1. Calculates great circle distances between storm center and control points
    2. Converts distances from degrees to kilometers
    3. Calculates angles relative to geographic north
    4. Transforms angles to relative coordinate system
    5. Adjusts angles for southern hemisphere
    """

    cp_dist_ls, cp_ang_ls = [], []
    for i in range(len(cp_lat_ls)):
        cp_lat, cp_lon = cp_lat_ls[i], cp_lon_ls[i]

        # get point polar reference
        # azimut is refered to geographical north (absolute system)
        arcl_h, ang_abs = geodesic_distance_azimuth(st_lat, st_lon, cp_lat, cp_lon)
        cp_dist_ls.append(arcl_h * np.pi / 180.0 * EARTH_RADIUS)  # [km]

        # change of coordinate system (absolute to relative)
        ang_rel = ang_abs - (aseg - 90)
        if ang_rel < 0:
            ang_rel = np.mod(ang_rel, 360)

        # south hemisphere effect
        if sign == -1:
            if (ang_rel >= 0) and (ang_rel <= 180):
                ang_rel = 180 - ang_rel
            elif (ang_rel >= 180) and (ang_rel <= 360):
                ang_rel = 360 - (ang_rel - 180)

        cp_ang_ls.append(ang_rel)

    return cp_dist_ls, cp_ang_ls  # [km], [º]


def get_mask_radii_angle(icase: int, mode: str = "") -> Tuple[np.ndarray, np.ndarray]:
    """
    Extract radii and angle indices for output points.

    Parameters
    ----------
    icase : int
        Analogue case ID.
    mode : str, optional
        Option to select SHyTCWaves library resolution ('', '_lowres'). Default is "".

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays containing:
        - rad : Radial distances from storm center (km)
        - ang : Angles from storm center (degrees)

    Notes
    -----
    The function:
    1. Loads output indices associated with distances/angles to target origin
    2. Determines grid size (small/medium/large) based on case ID
    3. Extracts appropriate radii and angle arrays for the grid size
    4. For low resolution mode, uses half the number of radial points
    """

    # load output indices associated with distances/angles to target origin
    if mode == "_lowres":
        xds_mask_ind = xr.open_dataset(PATHS["SHYTCWAVES_MDA_MASK_INDICES_LOWRES"])
    else:
        xds_mask_ind = xr.open_dataset(PATHS["SHYTCWAVES_MDA_MASK_INDICES"])

    # load MDA indices (grid sizes)
    xds_ind_mda = xr.open_dataset(PATHS["SHYTCWAVES_MDA_INDICES"])

    # get grid code
    pos_small = np.where(icase == xds_ind_mda.indices_small)[0]
    pos_medium = np.where(icase == xds_ind_mda.indices_medium)[0]
    pos_large = np.where(icase == xds_ind_mda.indices_large)[0]

    if len(pos_small) == 1:
        rad = xds_mask_ind.radii_sma.values[:] / 1000
        ang = xds_mask_ind.angle_sma.values[:]

    elif len(pos_medium) == 1:
        rad = xds_mask_ind.radii_med.values[:] / 1000
        ang = xds_mask_ind.angle_med.values[:]

    elif len(pos_large) == 1:
        rad = xds_mask_ind.radii_lar.values[:] / 1000
        ang = xds_mask_ind.angle_lar.values[:]

    return rad, ang  # [km], [º]


def find_nearest(
    cp_rad_ls: List[float],
    cp_ang_ls: List[float],
    mask_rad: np.ndarray,
    mask_ang: np.ndarray,
) -> np.ndarray:
    """
    Find nearest points in normalized space of radii and angles.

    Parameters
    ----------
    cp_rad_ls : List[float]
        Control point radial distances.
    cp_ang_ls : List[float]
        Control point angles.
    mask_rad : np.ndarray
        SWAN output point radial distances.
    mask_ang : np.ndarray
        SWAN output point angles.

    Returns
    -------
    np.ndarray
        Indices of nearest points in SWAN output grid for each control point.

    Notes
    -----
    The function:
    1. Creates dataframes from control points and SWAN grid points
    2. Treats radial distances as scalar values
    3. Treats angles as directional values (circular)
    4. Uses nearest neighbor search in normalized space
    """

    # create dataframes
    df_cp = pd.DataFrame({"radii": cp_rad_ls, "angle": cp_ang_ls})

    df_mask = pd.DataFrame({"radii": mask_rad, "angle": mask_ang})

    # indices
    ix_directional = [1]

    # get indices of nearest n-dimensional point
    ix_near = find_nearest_indices(
        query_points=df_cp.values,
        reference_points=df_mask.values,
        directional_indices=ix_directional,
    )

    return ix_near


###############################################################################
# SHyTCWaves APPLICATION

# Functions that calculate the ensemble/reconstruction for a storm track
# from either historical or forecast/predicted tracks
###############################################################################


def get_coef_calibration() -> np.ndarray:
    """
    Get calibration coefficients for SHyTCWaves model pressure bias correction.

    Returns
    -------
    np.ndarray
        Linear fit coefficients for pressure bias correction.

    Notes
    -----
    The function:
    1. Uses Saffir-Simpson category center pressures as reference points
    2. Applies calibrated pressure deltas based on validation against satellite data
    3. Performs linear fit to get correction coefficients
    """

    p = [1015, 990, 972, 954, 932, 880]  # Saffir-Simpson center categories
    dp = [-17, -15, -12.5, -7, +2.5, +10]  # calibrated "dP" for shytcwaves

    coef = np.polyfit(p, dp, 1)  # order 1 fitting

    return coef


##########################################
# SHYTCWAVES - historical track


def historic2shytcwaves_cluster(
    path_save: str,
    tc_name: str,
    storm: xr.Dataset,
    center: str,
    lon: np.ndarray,
    lat: np.ndarray,
    dict_site: Optional[dict] = None,
    calibration: bool = True,
    mode: str = "",
    database_on: bool = False,
    st_param: bool = False,
    extract_bulk: bool = True,
    max_segments: int = 300,
) -> None:
    """
    Process historical storm track data using SHyTCWaves methodology.

    Parameters
    ----------
    path_save : str
        Base path to store results, without the file name.
    tc_name : str
        Storm name.
    storm : xr.Dataset
        Storm track dataset with standard IBTrACS variables:

        Required:
        - longitude, latitude : Storm coordinates
        - pressure : Central pressure (mbar)
        - maxwinds : Maximum sustained winds (kt)

        Optional:
        - rmw : Radius of maximum winds (nmile)
        - dist2land : Distance to land
        - basin : Basin identifier

    center : str
        IBTrACS center code.
    lon : np.ndarray
        Longitude coordinates for swath calculation.
    lat : np.ndarray
        Latitude coordinates for swath calculation.
    dict_site : dict, optional
        Site data for superpoint building. Default is None.
        Must contain:
        - lonpts : Longitude coordinates
        - latpts : Latitude coordinates
        - namepts : Site names
        - site : Site identifier
        - sectors : Sectors
        - deg_superposition : Superposition degree

    calibration : bool, optional
        Whether to apply SHyTCWaves calibration. Default is True.
    mode : str, optional
        High or low resolution library indices. Default is "".
    database_on : bool, optional
        Whether to keep data only at 0,6,12,18 hours. Default is False.
    st_param : bool, optional
        Whether to keep data as original. Default is False.
    extract_bulk : bool, optional
        Whether to extract bulk wave parameters. Default is True.
    max_segments : int, optional
        Maximum number of segments to process. Default is 300.

    Notes
    -----
    The function processes historical storm tracks in several steps:
    1. Performs stopmotion segmentation at 6h intervals
    2. Optionally applies SHyTCWaves calibration to track parameters
    3. Trims track to target domain
    4. Finds analogue segments from library
    5. Extracts bulk wave parameters and/or spectral data
    6. Saves results to specified directory
    """

    # stopmotion segmentation, 6h interval
    df = historic_track_preprocessing(
        xds=storm,
        center=center,
        forecast_on=False,
        database_on=database_on,
        st_param=st_param,
    )
    dt_int_minutes = 6 * 60  # [minutes] constant segments

    # optional: shytcwaves calibration of track parameters
    if calibration:
        coef = get_coef_calibration()  # linear fitting
        df["pressure"] = df["pressure"].values * (1 + coef[0]) + coef[1]
        df["maxwinds"] = np.nan

        st, _ = historic_track_interpolation(
            df,
            dt_int_minutes,
            interpolation=False,
            mode="mean",
            fit=True,
            radi_estimate_on=True,
        )
    else:
        st, _ = historic_track_interpolation(
            df, dt_int_minutes, interpolation=False, mode="mean"
        )

    # skip when only NaN or 0
    lons, lats = st.lon.values, st.lat.values
    if (np.unique(lons[~np.isnan(lons)]).all() == 0) & (
        np.unique(lats[~np.isnan(lats)]).all() == 0
    ):
        print("No track coordinates")

    else:
        st_trim = track_triming(st, lat[0], lon[0], lat[-1], lon[-1])

        # store tracks for shytcwaves
        st.to_pickle(op.join(path_save, f"{tc_name}_track.pkl"))
        st_trim.to_pickle(op.join(path_save, f"{tc_name}_track_trim.pkl"))

        # parameterized segts (24h warmup + 6htarget)
        df_seg = storm2stopmotion(st_trim)

        if df_seg.shape[0] > 2:
            print(f"st: {st.shape[0]}, df_seg: {df_seg.shape[0]}")
            st_list, we_list = stopmotion_interpolation(df_seg, st=st_trim)

            # analogue segments from library
            df_mda = xr.open_dataset(PATHS["SHYTCWAVES_MDA"]).to_dataframe()
            ix_weights = [1] * 10  # equal weights
            ix = find_analogue(df_mda, df_seg, ix_weights)

            df_analogue = df_mda.iloc[ix]
            df_analogue = analogue_endpoints(df_seg, df_analogue)

            # extract bulk envelope (to plot swaths)
            if extract_bulk:
                mesh_lo, mesh_la = np.meshgrid(lon, lat)
                print(
                    f"Number of segments: {len(st_list)}, number of swath nodes: {mesh_lo.size}"
                )

                if len(st_list) < max_segments:
                    xds_shy_bulk = stopmotion_st_bmu(
                        df_analogue,
                        df_seg,
                        st_trim,
                        list(np.ravel(mesh_lo)),
                        list(np.ravel(mesh_la)),
                        max_dist=60,
                        mode=mode,
                    )
                    # store
                    xds_shy_bulk.to_netcdf(
                        op.join(path_save, f"{tc_name}_xds_shy_bulk.nc")
                    )

            # extract spectra envelope
            if isinstance(dict_site, dict):
                xds_shy_spec, _ = stopmotion_st_spectra(
                    df_analogue,
                    df_seg,
                    st_trim,
                    cp_lon_ls=dict_site["lonpts"],
                    cp_lat_ls=dict_site["latpts"],
                    cp_names=dict_site["namepts"],
                    mode=mode,
                )
                # store
                xds_shy_spec.to_netcdf(
                    op.join(
                        path_save,
                        f"{tc_name}_xds_shy_spec_{dict_site['site']}.nc",
                    )
                )

                # build superpoint
                xds_shy_sp = superpoint_calculation(
                    xds_shy_spec.efth,
                    "point",
                    dict_site["sectors"],
                )
                # store
                xds_shy_sp.to_netcdf(
                    op.join(
                        path_save,
                        f"{tc_name}_xds_shy_sp_{dict_site['site']}.nc",
                    )
                )

    print("Files stored.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

# import constants
from .config import _faspect, _fsize, _fdpi


def axplot_scatter_mda_vs_data(ax, x_mda, y_mda, x_data, y_data):
    'axes scatter plot variable1 vs variable2 mda vs data'

    # full dataset 
    ax.scatter(
        x_data, y_data,
        marker = '.',
        c = 'lightblue',
        s = 2, label = 'dataset'
    )

    # mda selection
    ax.scatter(
        x_mda, y_mda,
        marker = '.',
        c = 'k',
        s = 6, label='subset'
    )

def Plot_MDA_Data(pd_data, pd_mda, show=True):
    '''
    Plot scatter with MDA selection vs original data

    pd_data - pandas.DataFrame, complete data
    pd_mda - pandas.DataFrame, mda selected data

    pd_data and pd_mda should share columns names
    '''

    # TODO: activate dlab ? 

    ## figure conf.
    #d_lab = {
    #    'pressure_min': 'Pmin (mbar)',
    #    'gamma': 'gamma (º)',
    #    'delta': 'delta (º)',
    #    'velocity_mean': 'Vmean (km/h)',
    #}

    # variables to plot
    vns = pd_data.columns

    # filter
    vfs = ['n_sim']
    vns = [v for v in vns if v not in vfs]
    n = len(vns)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _faspect*_fsize))
    gs = gridspec.GridSpec(n-1, n-1, wspace=0.2, hspace=0.2)

    for i in range(n):
        for j in range(i+1, n):

            # get variables to plot
            vn1 = vns[i]
            vn2 = vns[j]

            # mda and entire-dataset
            vv1_mda = pd_mda[vn1].values[:]
            vv2_mda = pd_mda[vn2].values[:]

            vv1_dat = pd_data[vn1].values[:]
            vv2_dat = pd_data[vn2].values[:]

            # scatter plot 
            ax = plt.subplot(gs[i, j-1])
            axplot_scatter_mda_vs_data(ax, vv2_mda, vv1_mda, vv2_dat, vv1_dat)

            # custom axes
            if j==i+1:
                ax.set_xlabel(
                    #d_lab[vn2],
                    vn2,
                    {'fontsize':10, 'fontweight':'bold'}
                )
            if j==i+1:
                ax.set_ylabel(
                    #d_lab[vn1],
                    vn1,
                    {'fontsize':10, 'fontweight':'bold'}
                )

            if i==0 and j==n-1:
                ax.legend()

    # show and return figure
    if show: plt.show()
    return fig





#%%
#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import xarray as xr
import numpy as np
from scipy.interpolate import griddata
import matplotlib.pyplot as plt

# DEV: override installed teslakit
import sys
sys.path.insert(0, op.join(os.path.abspath(''), '..', '..'))

# teslakit
from teslakit.database import Database
from teslakit.io.matlab import ReadMatfile
from teslakit.rbf import RBF_Interpolation


from teslakit.util.time_operations import DateConverter_Mat2Py


## Database and Site parameters
# %%
# --------------------------------------
# Teslakit database

p_data = r'/Users/anacrueda/Documents/Proyectos/TESLA/Teslakit/data'
db = Database(p_data)

# set site
db.SetSite('KWAJALEIN')

# %%
# --------------------------------------
# load data and set parameters
#  wave data reconstructed a 30m water depth

p_hist = r'/Users/anacrueda/Documents/Proyectos/SERDP/results_files/Historicos/'

p_dataset = op.join(p_hist, 'Reconstr_Hs_Tp_Dir_Level_NOtransectRoi_Historical_2016_p6_in30m_snell_sep.mat')
p_gow =  op.join(p_hist, 'KWA_historical_parameters_2016_sep.mat')

subset = ReadMatfile(p_dataset)
timeo = ReadMatfile(p_gow)['time']
# parse matlab datenum to datetime
time = DateConverter_Mat2Py(timeo)

#define reef characteristics
Rslope = 0.0505*np.ones(len(time))
Bslope = 0.1667*np.ones(len(time))
Rw = 250*np.ones(len(time))
Cf = 0.0105*np.ones(len(time))


# %%
# dataset with wave and reef characteristics

hs_lo2 = subset['Hs']/(1.5613*subset['Tp']**2)

xds_subset = xr.Dataset(
    {
        'hs': ('time', subset['Hs']),
        'tp': ('time', subset['Tp']),
        'dir': ('time', subset['Dir']),
        'level': ('time', subset['Level']),
        'hs_lo2': ('time', hs_lo2),
        'rslope': ('time', Rslope),
        'bslope': ('time', Bslope),
        'rwidth': ('time', Rw),
        'cf': ('time', Cf),


    },

    coords={'time': time}
)

#print(xds_subset)

# load min, max, of simulated reef and wave conditions and rbfs coefficients

p_rbf = r'/Users/anacrueda/Documents/Proyectos/SERDP/HyCreWWcode/RBF_coefficients/'

p_min = op.join(p_rbf, 'Min_from_simulations.mat')
p_max = op.join(p_rbf, 'Max_from_simulations.mat')

smin = ReadMatfile(p_min)
smax = ReadMatfile(p_max)

# data max and min
dl = np.column_stack([smin['minimum'], smax['maximum']])

print(dl)

d = {
    'hs': dl[1],
    'tp': dl[2],
    'level':dl[0],
    'rslope': dl[3],
    'bslope': dl[4],
    'rwidth': dl[5],
    'cf': dl[6],
    'hs_lo2': [0.005, 0.05],
}

for k in d.keys():
    print(k, d[k])

# %%
# input data quality check
#['level','hs','tp','rslope','bslope','rwidth','cf']:

# NICO
xds_f = xds_subset.copy()

for vn in xds_subset.variables:
    if vn in d.keys():

        # get limits from dictionary
        l, h = tuple(d[vn])

        # find positions inside limits
        xds_f[vn] = xds_subset[vn].where((xds_subset[vn] >= l) & (xds_subset[vn] <= h))

print(xds_f)

#%%

# Normalize reef and water level parameters data

xds_reef = xr.Dataset({

    'level': ('time', xds_f['level']),
    'rslope': ('time', xds_f['rslope']),
    'bslope': ('time', xds_f['bslope']),
    'rwidth': ('time', xds_f['rwidth']),
    'cf': ('time', xds_f['cf']),
},
    coords={'time': xds_f['time']}
)

print(xds_reef)

xds_n = xds_reef.copy()
for k in xds_reef.variables:
    if k in d.keys():
        v = xds_n[k].values[:]
        l, h = tuple(d[k])
        xds_n[k] =(('time',), (v-l)/(h-l))


print()
print(xds_n)

#%%

# Hycreeww
# 15 different wave conditions were considered for each water level and reef characteristics
# the runup is calculated and then it is linearly interpolated for the actual wave conditions.

ncases = 15
runup = []

# Build RBF target numpy array

target = np.column_stack(
     (xds_n['level'], xds_n['rslope'], xds_n['bslope'], xds_n['rwidth'], xds_n['cf'])
)


target_wavecon = np.column_stack(
    (xds_f['hs'],xds_f['tp'],xds_f['hs_lo2'])
)

#%%

f_tot = []
for j in range(0, ncases):
    p_coeff= op.join(p_rbf, 'Coeffs_Runup_Xbeach_test' + str(j+1) + '.mat')
    coeff = ReadMatfile(p_coeff)
    rbf_constant = coeff['coeffsave']['RBFConstant']
    rbf_coeff = coeff['coeffsave']['rbfcoeff']
    nodes = coeff['coeffsave']['x']
    x = target
    f_temp = RBF_Interpolation(rbf_constant, rbf_coeff, nodes,x.T)
    f_tot.append(f_temp)

print(f_tot)

#linear interpolation

hs=[1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5] #Wave heigth conditions used in RBF
hs_lo=[0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05 ] #Wave conditions used in RBF

#%%

print(np.shape(f_tot))
print(target_wavecon.shape)
to_z = np.array(f_tot)
print(to_z.shape)
RU = []
for j in range (0,len(target)):
    x=hs; y=hs_lo; z=to_z[:,j]
    hs_e=target_wavecon[j,0]
    #tp_e=target_wavecon[j,1]
    hs_lo_e=target_wavecon[j,2]
    vq = griddata((x,y),z,(hs_e,hs_lo_e),method='linear')
    RU.append(vq)

#%%
#print(RU)

# we need to add the level to the hycreww estimation of Ru to get the proper runup
xds_results = xr.Dataset(
    {
        'hs': ('time', xds_f['hs']),
        'tp': ('time', xds_f['tp']),
        'dir': ('time', xds_f['dir']),
        'level': ('time', xds_f['level']),
        'hs_lo2': ('time', xds_f['hs_lo2']),
        'rslope': ('time', xds_f['rslope']),
        'bslope': ('time', xds_f['bslope']),
        'rwidth': ('time', xds_f['rwidth']),
        'cf': ('time', xds_f['cf']),
        'runup': ('time', RU + xds_f['level']),


    },

    coords={'time': time}
)

print(xds_results)

plt.figure(figsize=(13, 3.5))
ax = plt.gca()
xds_results.runup.plot(ax=ax)
plt.show()


#%%

#print(xds_results)

#runup_nan = xds_results.where(np.isnan(xds_results.runup) == 'True', drop='True')

#print(runup_nan)


import re
import unittest

from siphon.catalog import TDSCatalog

from bluemath_tk.config.paths import (
    PATHS,
    get_catalog_folders,
    get_paths,
    get_thredds_catalog,
    update_paths,
)


def test_paths_dictionary_structure():
    """
    Test that PATHS dictionary has the expected structure and keys.
    """

    tc = unittest.TestCase()
    required_keys = ["SHYTCWAVES_COEFS", "SHYTCWAVES_BULK", "SHYTCWAVES_MDA"]

    # Check all required keys exist
    for key in required_keys:
        tc.assertIn(key, PATHS, f"Missing required key: {key}")

    # Check all paths are strings
    for key, value in PATHS.items():
        tc.assertIsInstance(value, str, f"Path for {key} is not a string")

    # Check paths contain expected components using regex
    patterns = {
        "SHYTCWAVES_COEFS": r".*GEOOCEAN.*coef.*",
        "SHYTCWAVES_BULK": r".*GEOOCEAN.*bulk.*",
        "SHYTCWAVES_MDA": r".*GEOOCEAN.*mda.*",
    }

    for key, pattern in patterns.items():
        tc.assertTrue(
            re.search(pattern, PATHS[key], re.IGNORECASE),
            f"Path for {key} does not match pattern {pattern}",
        )


def test_update_paths():
    """
    Test the update_paths function.
    """

    tc = unittest.TestCase()
    # Test updating with new path
    new_path = "/test/path/to/data.nc"
    update_paths({"TEST_PATH": new_path})

    # Check new path was added
    tc.assertIn("TEST_PATH", PATHS)
    tc.assertEqual(PATHS["TEST_PATH"], new_path)


def test_get_paths():
    """
    Test the get_paths function returns the correct dictionary.
    """

    tc = unittest.TestCase()
    paths = get_paths()
    tc.assertIsInstance(paths, dict)
    tc.assertEqual(paths, PATHS)


def test_get_thredds_catalog():
    """
    Test thredds catalog is available.
    """

    tc = unittest.TestCase()
    catalog = get_thredds_catalog()
    tc.assertIsInstance(catalog, TDSCatalog)


def test_get_catalog_folders():
    """
    Test getting catalog folders from Thredds server.
    """

    tc = unittest.TestCase()
    folders = get_catalog_folders()
    tc.assertIsInstance(folders, dict)
    tc.assertGreater(len(folders), 0, "No folders found in catalog")


def suite():
    """
    Create a test suite.
    """

    suite = unittest.TestSuite()
    suite.addTest(unittest.FunctionTestCase(test_paths_dictionary_structure))
    suite.addTest(unittest.FunctionTestCase(test_update_paths))
    suite.addTest(unittest.FunctionTestCase(test_get_paths))
    suite.addTest(unittest.FunctionTestCase(test_get_thredds_catalog))
    suite.addTest(unittest.FunctionTestCase(test_get_catalog_folders))

    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
from datetime import timedelta
import numpy as np
import xarray as xr
from scipy.stats import  gumbel_l, genextreme

from .util.time_operations import date2datenum as d2d

def FitGEV_KMA_Frechet(bmus, n_clusters, var):
    '''
    Returns stationary GEV/Gumbel_L params for KMA bmus and varible series

    bmus        - KMA bmus (time series of KMA centroids)
    n_clusters  - number of KMA clusters
    var         - time series of variable to fit to GEV/Gumbel_L

    returns np.array (n_clusters x parameters). parameters = (shape, loc, scale)
    for gumbel distributions shape value will be ~0 (0.0000000001)
    '''

    param_GEV = np.empty((n_clusters, 3))
    for i in range(n_clusters):
        c = i+1
        pos = np.where((bmus==c))[0]

        if len(pos) == 0:
            param_GEV[i,:] = [np.nan, np.nan, np.nan]

        else:

            # get variable at cluster position
            var_c = var[pos]
            var_c = var_c[~np.isnan(var_c)]

            # fit to Gumbel_l and get negative loglikelihood
            loc_gl, scale_gl = gumbel_l.fit(-var_c)
            theta_gl = (0.0000000001, -1*loc_gl, scale_gl)
            nLogL_gl = genextreme.nnlf(theta_gl, var_c)

            # fit to GEV and get negative loglikelihood
            c = -0.1
            shape_gev, loc_gev, scale_gev = genextreme.fit(var_c, c)
            theta_gev = (shape_gev, loc_gev, scale_gev)
            nLogL_gev = genextreme.nnlf(theta_gev, var_c)

            # store negative shape
            theta_gev_fix = (-shape_gev, loc_gev, scale_gev)

            # apply significance test if Frechet
            if shape_gev < 0:

                # TODO: cant replicate ML exact solution
                if nLogL_gl - nLogL_gev >= 1.92:
                    param_GEV[i,:] = list(theta_gev_fix)
                else:
                    param_GEV[i,:] = list(theta_gl)
            else:
                param_GEV[i,:] = list(theta_gev_fix)

    return param_GEV

def Smooth_GEV_Shape(cenEOFs, param):
    '''
    Smooth GEV shape parameter (for each KMA cluster) by promediation
    with neighbour EOFs centroids

    cenEOFs  - (n_clusters, n_features) KMA centroids
    param    - GEV shape parameter for each KMA cluster

    returns smoothed GEV shape parameter as a np.array (n_clusters)
    '''

    # number of clusters
    n_cs = cenEOFs.shape[0]

    # calculate distances (optimized)
    cenEOFs_b = cenEOFs.reshape(cenEOFs.shape[0], 1, cenEOFs.shape[1])
    D = np.sqrt(np.einsum('ijk, ijk->ij', cenEOFs-cenEOFs_b, cenEOFs-cenEOFs_b))
    np.fill_diagonal(D, np.nan)

    # sort distances matrix to find neighbours
    sort_ord = np.empty((n_cs, n_cs), dtype=int)
    D_sorted = np.empty((n_cs, n_cs))
    for i in range(n_cs):
        order = np.argsort(D[i,:])
        sort_ord[i,:] = order
        D_sorted[i,:] = D[i, order]

    # calculate smoothed parameter
    denom = np.sum(1/D_sorted[:,:4], axis=1)
    param_c = 0.5 * np.sum(np.column_stack(
        [
            param[:],
            param[sort_ord[:,:4]] * (1/D_sorted[:,:4])/denom[:,None]
        ]
    ), axis=1)

    return param_c

def ACOV(f, theta, x):
    '''
    Returns asyntotyc variance matrix using Fisher Information matrix inverse
    Generalized functions, parameters and data.

    f      - function to evaluate: GEV, GUMBELL, ...
    theta  - function parameters: for GEV (shape, location, scale)
    x      - data used for function evaluation

    Second derivative evaluation - variance and covariance
    dxx = (f(x+dt_x) - 2f(x) + f(x-dt_x)) / (dt_x**2)
    dxy = (f(x,y) - f(x-dt_x,y) - f(x,y-dt_y) + f(x-dt_x, u-dt_y)) / (dt_x*dt_y)
    '''

    # parameters differential
    pm = 0.00001
    params = np.asarray(theta)
    dt_p = pm * params

    # Fisher information matrix holder 
    ss = len(params)
    FI = np.ones((ss,ss)) * np.nan

    # TODO: evaluar f falla en algunos casos?? 
    if np.isinf(f(theta, x)):
        #print ('ACOV error: nLogL = Inf. {0}'.format(theta))
        return np.ones((ss,ss))*0.0001

    # variance and covariance
    for i in range(ss):

        # diferential parameter FI evaluation
        p1 = np.asarray(theta); p1[i] = p1[i] + dt_p[i]
        p2 = np.asarray(theta); p2[i] = p2[i] - dt_p[i]

        # variance
        FI[i,i] = (f(tuple(p1), x) - 2*f(theta,x) + f(tuple(p2), x))/(dt_p[i]**2)

        for j in range(i+1,ss):

            # diferential parameter FI evaluation
            p1 = np.asarray(theta); p1[i] = p1[i] - dt_p[i]
            p2 = np.asarray(theta); p2[j] = p2[j] - dt_p[j]
            p3 = np.asarray(theta); p3[i] = p3[i] - dt_p[i]; p3[j] = p3[j] - dt_p[j]

            # covariance
            cov = (f(theta,x) - f(tuple(p1),x) - f(tuple(p2),x) + f(tuple(p3),x)) \
                    / (dt_p[i]*dt_p[j])
            FI[i,j] = cov
            FI[j,i] = cov

    # asynptotic variance covariance matrix
    acov = np.linalg.inv(FI)

    return acov

def Peaks_Over_Threshold(xds, var_name, percentile=99, threshold=None,
                         window_days=3):
    '''

    Peaks Over Threshold methodology to find extreme values.

    xds         - xarray.Dataset (time dim)
    var_name    - variable to apply POT
    percentile  - if threshold not given, calculate it with this percentile
    threshold   - optional, threshold to apply POT
    window_days - minimum number of days between consecutive independent peaks

    returns xarray.Dataset ('time', ) vars: peaks, excedeence, area, and duration
    '''

    # TODO: refactor with times
    def to_hours(time_diff):

        # single
        if isinstance(time_diff, (timedelta, np.timedelta64)):
            if isinstance(time_diff, np.timedelta64):
                dt_h = time_diff / np.timedelta64(1, 'h')

            elif isinstance(time_diff, timedelta):
                dt_h = np.array(time_diff.total_seconds()/3600.0)

            return dt_h

        # else: array
        if isinstance(time_diff[0], np.timedelta64):
            dt_h = time_diff / np.timedelta64(1, 'h')

        elif isinstance(time_diff[0], timedelta):
            dt_h = np.array([x.total_seconds()/3600.0 for x in time_diff])

        return dt_h


    # get variable
    vv = xds[var_name].values[:]
    vt = xds['time'].values[:]

    # data delta time (hours)
    npdf = np.diff(vt)
    dt_h = to_hours(npdf)

    # threshold
    if threshold == None:
        threshold = np.percentile(vv, percentile)

    # consecutive peaks over threshold
    ind_mask = np.where(vv >= threshold, 1, 0)
    ind_dif = np.diff(ind_mask)

    # peaks start and end
    ind_ini = np.where(ind_dif == 1)[0] + 1
    ind_end = np.where(ind_dif == -1)[0] + 1

    # start and end corrections
    if ind_end[0]<ind_ini[0]: ind_ini = np.insert(ind_ini,0,0)
    if ind_end[-1]<ind_ini[-1]: ind_end = np.append(ind_end, len(ind_dif))

    # store variable max, time max, duration and area above threshold for each peak
    time_max, vv_max, area, durac = [], [], [], []

    for i, f in zip(ind_ini, ind_end):

        vv_max.append(np.max(vv[i:f]))

        area.append(np.sum(vv[i:f] * dt_h[i:f] - threshold))

        ind_max = np.argmax(vv[i:f])
        ind_time = list(range(i, f))
        time_max.append(vt[ind_time[ind_max]])

        durac.append(to_hours(vt[f] - vt[i]))


    # ensure independence between maxs
    ind_window_mask = np.where(to_hours(np.diff(time_max)) < 24*window_days, 1, 0)
    if np.any(ind_window_mask):

        # solve dependent events
        ind_dif = np.diff(ind_window_mask)

        # first event corrections 
        ind_dif = np.insert(ind_dif, 0, 0)
        if ind_dif[1] == -1: ind_dif[0] = 1

        ind_ini = np.where(ind_dif == 1)[0]
        ind_fin = np.where(ind_dif == -1)[0] + 1

        # Keep maximum of dependent events
        time_max_indep = []
        vv_max_indep = []
        durac_indep = []
        area_indep = []
        ind_delete = []
        for ind_i, ind_f in zip(ind_ini, ind_fin):

            vv_temp = np.max(vv_max[ind_i:ind_f])
            vv_max_indep.append(vv_temp)

            vv_ind_max = np.argmax(vv_max[ind_i:ind_f])
            ind_time = list(range(ind_i,ind_f))
            time_temp = time_max[ind_time[vv_ind_max]]
            time_max_indep.append(time_temp)

            durac_indep.append(
                to_hours(time_max[ind_f]-time_max[ind_i])
            )

            area_indep.append(np.sum(area[ind_i:ind_f]))

            ind_delete.extend(ind_time)

        # remove all dependent events
        vv_max = np.delete(vv_max, [ind_delete])
        time_max = np.delete(time_max, [ind_delete])
        durac = np.delete(durac, [ind_delete])
        area = np.delete(area, [ind_delete])

        # add independent events
        vv_max = np.concatenate((vv_max, vv_max_indep))
        time_max = np.concatenate((time_max, time_max_indep))
        durac = np.concatenate((durac, durac_indep))  # hours
        area = np.concatenate((area, area_indep))

    # output
    peaks = xr.Dataset(
        {
            '{0}'.format(var_name): (('time',), vv_max),
            '{0}_exceedances'.format(var_name): (('time',), vv_max-threshold),
            'duration': (('time',), durac),
            'area': (('time',), area),
        },
        coords = {'time': time_max},
    )
    peaks = peaks.sortby('time')

    # add some attrs
    peaks.attrs['threshold'] = threshold
    peaks.attrs['percentile'] = percentile

    return peaks






from typing import Dict, List

import numpy as np

from ._base_distributions import BaseDistribution, FitResult, fit_dist


class GPD(BaseDistribution):
    """
    Generalized Pareto Distribution (GPD) class.

    This class contains all the methods assocaited to the GPD distribution.

    Attributes
    ----------
    name : str
        The complete name of the distribution (GPD).
    nparams : int
        Number of GPD parameters.
    param_names : List[str]
        Names of the GPD parameters (threshold, scale, shape).

    Methods
    -------
    pdf(x, loc, scale, shape)
        Probability density function.
    cdf(x, loc, scale, shape)
        Cumulative distribution function
    qf(p, loc, scale, shape)
        Quantile function
    sf(x, loc, scale, shape)
        Survival function
    nll(data, loc, scale, shape)
        Negative Log-Likelihood function
    fit(data)
        Fit distribution to data (NOT IMPLEMENTED).
    random(size, loc, scale, shape)
        Generates random values from GPD distribution.
    mean(loc, scale, shape)
        Mean of GPD distribution.
    median(loc, scale, shape)
        Median of GPD distribution.
    variance(loc, scale, shape)
        Variance of GPD distribution.
    std(loc, scale, shape)
        Standard deviation of GPD distribution.
    stats(loc, scale, shape)
        Summary statistics of GPD distribution.

    Notes
    -----
    - This class is designed to obtain all the properties associated to the GPD distribution.

    Examples
    --------
    >>> from bluemath_tk.distributions.gpd import GPD
    >>> gpd_pdf = GPD.pdf(x, loc=0, scale=1, shape=0.1)
    >>> gpd_cdf = GPD.cdf(x, loc=0, scale=1, shape=0.1)
    >>> gpd_qf = GPD.qf(p, loc=0, scale=1, shape=0.1)
    """

    def __init__(self) -> None:
        """
        Initialize the GPD distribution class
        """
        super().__init__()

    @staticmethod
    def name() -> str:
        return "Generalized Pareto Distribution (GPD)"

    @staticmethod
    def nparams() -> int:
        """
        Number of parameters of GPD
        """
        return int(3)

    @staticmethod
    def param_names() -> List[str]:
        """
        Name of parameters of GPD
        """
        return ["loc", "scale", "shape"]

    @staticmethod
    def pdf(
        x: np.ndarray, loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> np.ndarray:
        """
        Probability density function

        Parameters
        ----------
        x : np.ndarray
            Values to compute the probability density value
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        pdf : np.ndarray
            Probability density function values

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        y = np.maximum(x - loc, 0) / scale

        # Gumbel case (shape = 0)
        if shape == 0.0:
            pdf = (1 / scale) * (np.exp(-y))

        # General case (Weibull and Frechet, shape != 0)
        else:
            pdf = np.full_like(x, 0, dtype=float)
            yy = 1 + shape * y
            yymask = yy > 0
            pdf[yymask] = (1 / scale) * (yy[yymask] ** (-1 - (1 / shape)))

        return pdf

    @staticmethod
    def cdf(
        x: np.ndarray, loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> np.ndarray:
        """
        Cumulative distribution function

        Parameters
        ----------
        x : np.ndarray
            Values to compute their probability
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        p : np.ndarray
            Probability

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        y = np.maximum(x - loc, 0) / scale

        # Gumbel case (shape = 0)
        if shape == 0.0:
            p = 1 - np.exp(-y)

        # General case (Weibull and Frechet, shape != 0)
        else:
            p = 1 - np.maximum(1 + shape * y, 0) ** (-1 / shape)

        return p

    @staticmethod
    def sf(
        x: np.ndarray, loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> np.ndarray:
        """
        Survival function (1-Cumulative Distribution Function)

        Parameters
        ----------
        x : np.ndarray
            Values to compute their survival function value
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        sp : np.ndarray
            Survival function value

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        sp = 1 - GPD.cdf(x, loc=loc, scale=scale, shape=shape)

        return sp

    @staticmethod
    def qf(
        p: np.ndarray, loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> np.ndarray:
        """
        Quantile function (Inverse of Cumulative Distribution Function)

        Parameters
        ----------
        p : np.ndarray
            Probabilities to compute their quantile
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        q : np.ndarray
            Quantile value

        Raises
        ------
        ValueError
            If probabilities are not in the range (0, 1).

        ValueError
            If scale is not greater than 0.
        """

        if np.min(p) <= 0 or np.max(p) >= 1:
            raise ValueError("Probabilities must be in the range (0, 1)")

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        # Gumbel case (shape = 0)
        if shape == 0.0:
            q = loc - scale * np.log(1 - p)

        # General case (Weibull and Frechet, shape != 0)
        else:
            q = loc + scale * ((1 - p) ** (-shape) - 1) / shape

        return q

    @staticmethod
    def nll(
        data: np.ndarray, loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> float:
        """
        Negative Log-Likelihood function

        Parameters
        ----------
        data : np.ndarray
            Data to compute the Negative Log-Likelihood value
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        nll : float
            Negative Log-Likelihood value
        """

        if scale <= 0:
            nll = np.inf  # Return a large value for invalid scale

        else:
            y = (data - loc) / scale

            # # Gumbel case (shape = 0)
            # if shape == 0.0:
            #     nll = data.shape[0] * np.log(scale) + np.sum(y)

            # General case (Weibull and Frechet, shape != 0)
            # else:
            shape = (
                np.maximum(shape, 1e-8) if shape > 0 else np.minimum(shape, -1e-8)
            )  # Avoid division by zero
            y = 1 + shape * y
            if np.min(y <= 0):
                nll = np.inf  # Return a large value for invalid y
            else:
                nll = data.shape[0] * np.log(scale) + (1 / shape + 1) * np.sum(
                    np.log(y)
                )

        return nll

    @staticmethod
    def fit(data: np.ndarray, **kwargs) -> FitResult:
        """
        Fit GEV distribution

        Parameters
        ----------
        data : np.ndarray
            Data to fit the GEV distribution
        **kwargs : dict, optional
            Additional keyword arguments for the fitting function.
            These can include options like method, bounds, etc.
            See fit_dist for more details.
            If not provided, default fitting options will be used.

        Returns
        ----------
        FitResult
            Result of the fit containing the parameters loc, scale, shape,
            success status, and negative log-likelihood value.
        """
        # Fit the GEV distribution to the data using the fit_dist function
        return fit_dist(GPD, data, **kwargs)

    @staticmethod
    def random(
        size: int,
        loc: float = 0.0,
        scale: float = 1.0,
        shape: float = 0.0,
        random_state: int = None,
    ) -> np.ndarray:
        """
        Generates random values from GPD distribution

        Parameters
        ----------
        size : int
            Number of random values to generate
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.
        random_state : np.random.RandomState, optional
            Random state for reproducibility.
            If None, do not use random stat.

        Returns
        ----------
        x : np.ndarray
            Random values from GEV distribution

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        # Set random state if provided
        if random_state is not None:
            np.random.seed(random_state)

        # Generate uniform random numbers
        u = np.random.uniform(0, 1, size)

        # Gumbel case (shape = 0)
        if shape == 0.0:
            x = loc - scale * np.log(u)

        # General case (Weibull and Frechet, shape != 0)
        else:
            x = loc + scale * (u ** (-shape) - 1) / shape

        return x

    @staticmethod
    def mean(loc: float = 0.0, scale: float = 1.0, shape: float = 0.0) -> float:
        """
        Mean

        Parameters
        ----------
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        mean : np.ndarray
            Mean value of GEV with the given parameters

        Raises
        ------
        ValueError
            If scale is not greater than 0.

        Warning
            If shape is greater than or equal to 1, mean is not defined.
            In this case, it returns infinity.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        if shape >= 1:
            Warning("Shape parameter must be < 1 for mean to be defined")
            mean = np.inf

        # Shape < 1 case
        else:
            mean = scale / (1 - shape)

        return mean

    @staticmethod
    def median(loc: float = 0.0, scale: float = 1.0, shape: float = 0.0) -> float:
        """
        Median

        Parameters
        ----------
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        median : np.ndarray
            Median value of GEV with the given parameters

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        if shape == 0:
            median = np.inf
        else:
            median = loc + scale * (2**shape - 1) / shape

        return median

    @staticmethod
    def variance(loc: float = 0.0, scale: float = 1.0, shape: float = 0.0) -> float:
        """
        Variance

        Parameters
        ----------
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        var : np.ndarray
            Variance of GEV with the given parameters

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        Warning
            If shape is greater than or equal to 172, mean is not defined.
            In this case, it returns infinity.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        # Gumbel case (shape = 0)
        if shape >= 1 / 2:
            Warning("Shape parameter must be < 1/2 for variance to be defined")
            var = np.inf

        else:
            var = scale**2 / ((1 - shape) ** 2 * (1 - 2 * shape))

        return var

    @staticmethod
    def std(loc: float = 0.0, scale: float = 1.0, shape: float = 0.0) -> float:
        """
        Standard deviation

        Parameters
        ----------
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        std : np.ndarray
            Standard Deviation of GEV with the given
            parameters

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        std = np.sqrt(GPD.variance(loc, scale, shape))

        return std

    @staticmethod
    def stats(
        loc: float = 0.0, scale: float = 1.0, shape: float = 0.0
    ) -> Dict[str, float]:
        """
        Summary statistics

        Return summary statistics including mean, std, variance, etc.

        Parameters
        ----------
        loc : float, default=0.0
            Location parameter
        scale : float, default = 1.0
            Scale parameter.
            Must be greater than 0.
        shape : float, default = 0.0
            Shape parameter.

        Returns
        ----------
        stats : dict
            Summary statistics of GEV distribution with the given
            parameters

        Raises
        ------
        ValueError
            If scale is not greater than 0.
        """

        if scale <= 0:
            raise ValueError("Scale parameter must be > 0")

        stats = {
            "mean": float(GPD.mean(loc, scale, shape)),
            "median": float(GPD.median(loc, scale, shape)),
            "variance": float(GPD.variance(loc, scale, shape)),
            "std": float(GPD.std(loc, scale, shape)),
        }

        return stats



from abc import abstractmethod
from typing import List, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..core.models import BlueMathModel
from ..core.plotting.base_plotting import DefaultStaticPlotting
from ..core.plotting.scatter import plot_scatters_in_triangle
from ..core.plotting.utils import get_list_of_colors_for_colormap


class BaseSampling(BlueMathModel):
    """
    Base class for all sampling BlueMath models.
    This class provides the basic structure for all sampling models.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def generate(self, *args, **kwargs) -> pd.DataFrame:
        """
        Generates samples.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            The generated samples.
        """

        return pd.DataFrame()

    def plot_generated_data(
        self,
        data_color: str = "blue",
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        """
        Plots the generated data on a scatter plot matrix.

        Parameters
        ----------
        data_color : str, optional
            Color for the data points. Default is "blue".
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        Figure
            The figure object containing the plot.
        Axes
            Array of axes objects for the subplots.

        Raises
        ------
        ValueError
            If the data is empty.
        """

        if self.data.empty:
            raise ValueError("Data must be a non-empty DataFrame with columns to plot.")

        fig, axes = plot_scatters_in_triangle(
            dataframes=[self.data],
            data_colors=[data_color],
            **kwargs,
        )

        return fig, axes


class BaseClustering(BlueMathModel):
    """
    Base class for all clustering BlueMath models.
    This class provides the basic structure for all clustering models.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()

        self._exclude_attributes = [
            "_data",
            "_normalized_data",
            "_data_to_fit",
        ]

    @abstractmethod
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        normalize_data: bool = False,
    ) -> None:
        """
        Preprocess some data to be used in the fit of children classes.
        """

        self._data = data.copy()
        self.directional_variables = directional_variables.copy()
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=self.data[directional_variable].values
            )
            self.data[f"{directional_variable}_u"] = u_comp
            self.data[f"{directional_variable}_v"] = v_comp
        self.data_variables = list(self.data.columns)

        # Get just the data to be used in the training
        self._data_to_fit = self.data.copy()
        for directional_variable in self.directional_variables:
            self.data_to_fit.drop(columns=[directional_variable], inplace=True)
        self.fitting_variables = list(self.data_to_fit.columns)

        if normalize_data:
            self.custom_scale_factor = custom_scale_factor.copy()
        else:
            self.logger.info(
                "Normalization is disabled. Set normalize_data to True to enable normalization."
            )
            self.custom_scale_factor = {
                fitting_variable: (0, 1) for fitting_variable in self.fitting_variables
            }
        # Normalize data using custom min max scaler
        self._normalized_data, self.scale_factor = self.normalize(
            data=self.data_to_fit, custom_scale_factor=self.custom_scale_factor
        )

    @abstractmethod
    def predict(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Preprocess some data to be used in the predict of children classes.
        """

        data = data.copy()  # Avoid modifying the original data to predict
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=data[directional_variable].values
            )
            data[f"{directional_variable}_u"] = u_comp
            data[f"{directional_variable}_v"] = v_comp
            data.drop(columns=[directional_variable], inplace=True)
        normalized_data, _ = self.normalize(
            data=data, custom_scale_factor=self.scale_factor
        )

        return normalized_data

    def plot_selected_centroids(
        self,
        data_color: str = "blue",
        centroids_color: str = "red",
        plot_text: bool = False,
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        """
        Plots data and selected centroids on a scatter plot matrix.

        Parameters
        ----------
        data_color : str, optional
            Color for the data points. Default is "blue".
        centroids_color : str, optional
            Color for the centroid points. Default is "red".
        plot_text : bool, optional
            Whether to display text labels for centroids. Default is False.
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        Figure
            The figure object containing the plot.
        Axes
            Array of axes objects for the subplots.

        Raises
        ------
        ValueError
            If the data and centroids do not have the same number of columns or if the columns are empty.
        """

        if (
            len(self.data.columns) == len(self.centroids.columns)
            and list(self.data.columns) != []
        ):
            variables_names = list(self.data.columns)
        else:
            raise ValueError(
                "Data and centroids must have the same number of columns > 0."
            )

        fig, axes = plot_scatters_in_triangle(
            dataframes=[self.data, self.centroids],
            data_colors=[data_color, centroids_color],
            **kwargs,
        )
        if plot_text:
            for c1, v1 in enumerate(variables_names[1:]):
                for c2, v2 in enumerate(variables_names[:-1]):
                    for i in range(self.centroids.shape[0]):
                        axes[c2, c1].text(
                            self.centroids[v1][i],
                            self.centroids[v2][i],
                            str(i + 1),
                            fontsize=12,
                            fontweight="bold",
                        )

        return fig, axes

    def plot_data_as_clusters(
        self,
        data: pd.DataFrame,
        nearest_centroids: np.ndarray,
        **kwargs,
    ) -> Tuple[Figure, Axes]:
        """
        Plots data as nearest clusters.

        Parameters
        ----------
        data : pd.DataFrame
            The data to plot.
        nearest_centroids : np.ndarray
            The nearest centroids.
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the scatter plot function.

        Returns
        -------
        Figure
            The figure object containing the plot.
        Axes
            The axes object for the plot.
        """

        if (
            not data.empty
            and list(self.data.columns) != []
            and nearest_centroids.size > 0
        ):
            variables_names = list(data.columns)
            num_variables = len(variables_names)
        else:
            raise ValueError(
                "Data must have columns and nearest centroids must have values."
            )

        # Create figure and axes
        default_static_plot = DefaultStaticPlotting()
        fig, axes = default_static_plot.get_subplots(
            nrows=num_variables - 1,
            ncols=num_variables - 1,
            sharex=False,
            sharey=False,
        )
        if isinstance(axes, Axes):
            axes = np.array([[axes]])

        # Gets colors for clusters and append to each nearest centroid
        colors_for_clusters = get_list_of_colors_for_colormap(
            cmap="jet", num_colors=self.centroids.shape[0]
        )
        nearest_centroids_colors = [colors_for_clusters[i] for i in nearest_centroids]

        for c1, v1 in enumerate(variables_names[1:]):
            for c2, v2 in enumerate(variables_names[:-1]):
                default_static_plot.plot_scatter(
                    ax=axes[c2, c1],
                    x=data[v1],
                    y=data[v2],
                    c=nearest_centroids_colors,
                    alpha=0.9,
                    **kwargs,
                )
                if c1 == c2:
                    axes[c2, c1].set_xlabel(variables_names[c1 + 1])
                    axes[c2, c1].set_ylabel(variables_names[c2])
                elif c1 > c2:
                    axes[c2, c1].xaxis.set_ticklabels([])
                    axes[c2, c1].yaxis.set_ticklabels([])
                else:
                    fig.delaxes(axes[c2, c1])

        return fig, axes


class BaseReduction(BlueMathModel):
    """
    Base class for all dimensionality reduction BlueMath models.
    This class provides the basic structure for all dimensionality reduction models.
    """

    @abstractmethod
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def fit(self, *args, **kwargs) -> None:
        """
        Fits the model to the data.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        pass

    @abstractmethod
    def transform(self, *args, **kwargs) -> xr.Dataset:
        """
        Transforms the data using the fitted model.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        xr.Dataset
            The transformed data.
        """

        return xr.Dataset()

    @abstractmethod
    def fit_transform(self, *args, **kwargs) -> xr.Dataset:
        """
        Fits the model to the data and transforms it.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        xr.Dataset
            The transformed data.
        """

        return xr.Dataset()

    @abstractmethod
    def inverse_transform(self, *args, **kwargs) -> xr.Dataset:
        """
        Inversely transforms the data using the fitted model.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        xr.Dataset
            The inversely transformed data.
        """

        return xr.Dataset()


class ClusteringComparator:
    """
    Class for comparing clustering models.
    """

    def __init__(self, list_of_models: List[BaseClustering]) -> None:
        """
        Initializes the ClusteringComparator class.
        """

        self.list_of_models = list_of_models

    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
    ) -> None:
        """
        Fits the clustering models.
        """

        for model in self.list_of_models:
            if model.__class__.__name__ == "SOM":
                model.fit(
                    data=data,
                    directional_variables=directional_variables,
                )
            else:
                model.fit(
                    data=data,
                    directional_variables=directional_variables,
                    custom_scale_factor=custom_scale_factor,
                )

    def plot_selected_centroids(self) -> None:
        """
        Plots the selected centroids for the clustering models.
        """

        for model in self.list_of_models:
            fig, axes = model.plot_selected_centroids()
            fig.suptitle(f"Selected centroids for {model.__class__.__name__}")

    def plot_data_as_clusters(self, data: pd.DataFrame) -> None:
        """
        Plots the data as clusters for the clustering models.
        """

        for model in self.list_of_models:
            nearest_centroids, _ = model.predict(data=data)
            fig, axes = model.plot_data_as_clusters(
                data=data, nearest_centroids=nearest_centroids
            )
            fig.suptitle(f"Data as clusters for {model.__class__.__name__}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# pip
import numpy as np
import statsmodels.api as sm
from statsmodels.distributions.empirical_distribution import ECDF
from scipy.interpolate import interp1d
from scipy.stats import norm, genpareto, t
from scipy.special import ndtri  # norm inv


def ksdensity_CDF(x):
    '''
    Kernel smoothing function estimate.
    Returns cumulative probability function at x.
    '''

    # fit a univariate KDE
    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit()

    # interpolate KDE CDF at x position (kde.support = x) 
    fint = interp1d(kde.support, kde.cdf)

    return fint(x)

def ksdensity_ICDF(x, p):
    '''
    Returns Inverse Kernel smoothing function at p points
    '''

    # fit a univariate KDE
    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit()

    # interpolate KDE CDF to get support values 
    fint = interp1d(kde.cdf, kde.support)

    # ensure p inside kde.cdf
    p[p<np.min(kde.cdf)] = kde.cdf[0]
    p[p>np.max(kde.cdf)] = kde.cdf[-1]

    return fint(p)

def GeneralizedPareto_CDF(x):
    '''
    Generalized Pareto fit
    Returns cumulative probability function at x.
    '''

    # fit a generalized pareto and get params 
    shape, _, scale = genpareto.fit(x)

    # get generalized pareto CDF
    cdf = genpareto.cdf(x, shape, scale=scale)

    return cdf

def GeneralizedPareto_ICDF(x, p):
    '''
    Generalized Pareto fit
    Returns inverse cumulative probability function at p points
    '''

    # fit a generalized pareto and get params 
    shape, _, scale = genpareto.fit(x)

    # get percent points (inverse of CDF) 
    icdf = genpareto.ppf(p, shape, scale=scale)

    return icdf

def Empirical_CDF(x):
    '''
    Returns empirical cumulative probability function at x.
    '''

    # fit ECDF
    ecdf = ECDF(x)
    cdf = ecdf(x)

    return cdf

def Empirical_ICDF(x, p):
    '''
    Returns inverse empirical cumulative probability function at p points
    '''

    # TODO: revisar que el fill_value funcione correctamente

    # fit ECDF
    ecdf = ECDF(x)
    cdf = ecdf(x)

    # interpolate KDE CDF to get support values 
    fint = interp1d(
        cdf, x,
        fill_value=(np.nanmin(x), np.nanmax(x)),
        #fill_value=(np.min(x), np.max(x)),
        bounds_error=False
    )
    return fint(p)

def copulafit(u, family='gaussian'):
    '''
    Fit copula to data.
    Returns correlation matrix and degrees of freedom for t student
    '''

    rhohat = None  # correlation matrix
    nuhat = None  # degrees of freedom (for t student) 

    if family=='gaussian':
        u[u>=1.0] = 0.999999
        inv_n = ndtri(u)
        rhohat = np.corrcoef(inv_n.T)

    elif family=='t':
        raise ValueError("Not implemented")

        # TODO: no encaja con los datos. no funciona 
        x = np.linspace(np.min(u), np.max(u),100)
        inv_t = np.ndarray((len(x), u.shape[1]))

        for j in range(u.shape[1]):
            param = t.fit(u[:,j])
            t_pdf = t.pdf(x,loc=param[0],scale=param[1],df=param[2])
            inv_t[:,j] = t_pdf

        # TODO CORRELACION? NUHAT?
        rhohat = np.corrcoef(inv_n.T)
        nuhat = None

    else:
        raise ValueError("Wrong family parameter. Use 'gaussian' or 't'")

    return rhohat, nuhat

def copularnd(family, rhohat, n):
    '''
    Random vectors from a copula
    '''

    if family=='gaussian':
        mn = np.zeros(rhohat.shape[0])
        np_rmn = np.random.multivariate_normal(mn, rhohat, n)
        u = norm.cdf(np_rmn)

    elif family=='t':
        # TODO
        raise ValueError("Not implemented")

    else:
        raise ValueError("Wrong family parameter. Use 'gaussian' or 't'")

    return u

def CopulaSimulation(U_data, kernels, num_sim):
    '''
    Fill statistical space using copula simulation

    U_data: 2D nump.array, each variable in a column
    kernels: list of kernels for each column at U_data (KDE | GPareto)
    num_sim: number of simulations
    '''

    # kernel CDF dictionary
    d_kf = {
        'KDE' : (ksdensity_CDF, ksdensity_ICDF),
        'GPareto' : (GeneralizedPareto_CDF, GeneralizedPareto_ICDF),
        'ECDF' : (Empirical_CDF, Empirical_ICDF),
    }

    # check kernel input
    if any([k not in d_kf.keys() for k in kernels]):
        raise ValueError(
            'wrong kernel: {0}, use: {1}'.format(
                kernel, ' | '.join(d_kf.keys())
            )
        )

    # normalize: calculate data CDF using kernels
    U_cdf = np.zeros(U_data.shape) * np.nan
    ic = 0
    for d, k in zip(U_data.T, kernels):
        cdf, _ = d_kf[k]  # get kernel cdf
        U_cdf[:, ic] = cdf(d)
        ic += 1

    # fit data CDFs to a gaussian copula 
    rhohat, _ = copulafit(U_cdf, 'gaussian')

    # simulate data to fill probabilistic space
    U_cop = copularnd('gaussian', rhohat, num_sim)

    # de-normalize: calculate data ICDF
    U_sim = np.zeros(U_cop.shape) * np.nan
    ic = 0
    for d, c, k in zip(U_data.T, U_cop.T, kernels):
        _, icdf = d_kf[k]  # get kernel icdf
        U_sim[:, ic] = icdf(d, c)
        ic += 1

    return U_sim

def runmean(X, m, modestr):
    '''
    parsed runmean function from original matlab codes.
    '''

    mm = 2*m+1

    if modestr == 'edge':
        xfirst = np.repeat(X[0], m)
        xlast = np.repeat(X[-1], m)
    elif modestr == 'zero':
        xfirst = np.zeros(m)
        xlast = np.zeros(m)
    elif modestr == 'mean':
        xfirst = np.repeat(np.nanmean(X), m)
        xlast = xfirst

    Y = np.concatenate(
        (np.zeros(1), xfirst, X, xlast)
    )
    Y = np.nancumsum(Y)
    Y = np.divide(Y[mm:,]-Y[:-mm], mm)

    return Y




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
from datetime import datetime

# pip
import numpy as np
import xarray as xr
from sklearn.decomposition import PCA


def running_mean(x, N, mode_str='mean'):
    '''
    computes a running mean (also known as moving average)
    on the elements of the vector X. It uses a window of 2*M+1 datapoints

    As always with filtering, the values of Y can be inaccurate at the
    edges. RUNMEAN(..., MODESTR) determines how the edges are treated. MODESTR can be
    one of the following strings:
      'edge'    : X is padded with first and last values along dimension
                  DIM (default)
      'zeros'   : X is padded with zeros
      'ones'    : X is padded with ones
      'mean'    : X is padded with the mean along dimension DIM

    X should not contains NaNs, yielding an all NaN result.
    '''

    # if nan in data, return nan array
    if np.isnan(x).any():
        return np.full(x.shape, np.nan)

    nn = 2*N+1

    if mode_str == 'zeros':
        x = np.insert(x, 0, np.zeros(N))
        x = np.append(x, np.zeros(N))

    elif mode_str == 'ones':
        x = np.insert(x, 0, np.ones(N))
        x = np.append(x, np.ones(N))

    elif mode_str == 'edge':
        x = np.insert(x, 0, np.ones(N)*x[0])
        x = np.append(x, np.ones(N)*x[-1])

    elif mode_str == 'mean':
        x = np.insert(x, 0, np.ones(N)*np.mean(x))
        x = np.append(x, np.ones(N)*np.mean(x))


    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[nn:] - cumsum[:-nn]) / float(nn)

def RunnningMean_Monthly(xds, var_name, window=5):
    '''
    Calculate running average grouped by months

    xds:
        (longitude, latitude, time) variables: var_name

    returns xds with new variable "var_name_runavg"
    '''

    tempdata_runavg = np.empty(xds[var_name].shape)

    for lon in xds.longitude.values:
       for lat in xds.latitude.values:
          for mn in range(1, 13):

             # indexes
             ix_lon = np.where(xds.longitude == lon)
             ix_lat = np.where(xds.latitude == lat)
             ix_mnt = np.where(xds['time.month'] == mn)

             # point running average
             time_mnt = xds.time[ix_mnt]
             data_pnt = xds[var_name].loc[lon, lat, time_mnt]

             tempdata_runavg[ix_lon[0], ix_lat[0], ix_mnt[0]] = running_mean(
                 data_pnt.values, window)

    # store running average
    xds['{0}_runavg'.format(var_name)]= (
        ('longitude', 'latitude', 'time'),
        tempdata_runavg)

    return xds

def PCA_LatitudeAverage(xds, var_name, y1, y2, m1, m2):
    '''
    Principal component analysis
    method: remove monthly running mean and latitude average

    xds:
        (longitude, latitude, time), pred_name | pred_name_runavg

    returns a xarray.Dataset containing PCA data: PCs, EOFs, variance
    '''

    # calculate monthly running mean
    xds = RunnningMean_Monthly(xds, var_name)

    # predictor variable and variable_runnavg from dataset
    var_val = xds[var_name]
    var_val_ra = xds['{0}_runavg'.format(var_name)]

    # use datetime for indexing
    dt1 = datetime(y1, m1, 1)
    dt2 = datetime(y2+1, m2, 28)
    time_PCA = [datetime(y, m1, 1) for y in range(y1, y2+1)]

    # use data inside timeframe
    data_ss = var_val.loc[:,:,dt1:dt2]
    data_ss_ra = var_val_ra.loc[:,:,dt1:dt2]

    # anomalies: remove the monthly running mean
    data_anom = data_ss - data_ss_ra

    # average across all latitudes
    data_avg_lat = data_anom.mean(dim='latitude')

    # collapse 12 months of data to a single vector
    nlon = data_avg_lat.longitude.shape[0]
    ntime = data_avg_lat.time.shape[0]
    hovmoller = xr.DataArray(
        np.reshape(data_avg_lat.values, (12*nlon, ntime//12), order='F')
    )
    hovmoller = hovmoller.transpose()

    # mean and standard deviation
    var_anom_mean = hovmoller.mean(axis=0)
    var_anom_std = hovmoller.std(axis=0)

    # remove means and normalize by the standard deviation at anomaly
    # rows = time, columns = longitude
    nk_m = np.kron(np.ones((y2-y1+1,1)), var_anom_mean)
    nk_s = np.kron(np.ones((y2-y1+1,1)), var_anom_std)
    var_anom_demean = (hovmoller - nk_m) / nk_s

    # sklearn principal components analysis
    ipca = PCA(n_components=var_anom_demean.shape[0])
    PCs = ipca.fit_transform(var_anom_demean)

    pred_lon = xds.longitude.values[:]

    return xr.Dataset(
        {
            'PCs': (('n_components', 'n_components'), PCs),
            'EOFs': (('n_components','n_features'), ipca.components_),
            'variance': (('n_components',), ipca.explained_variance_),

            'var_anom_std': (('n_features',), var_anom_std),
            'var_anom_mean': (('n_features',), var_anom_mean),

            'time': (('n_components'), time_PCA),
            'pred_lon': (('n_lon',), pred_lon),
        },

        # store PCA algorithm metadata
        attrs = {
            'method': 'anomalies, latitude averaged',
        }
    )

def PCA_EstelaPred(xds, pred_name):
    '''
    Principal component analysis
    method: custom for estela predictor

    xds:
        (time, latitude, longitude), pred_name_comp | pred_name_gradient_comp

    returns a xarray.Dataset containing PCA data: PCs, EOFs, variance
    '''

    # estela predictor and estela gradient predictor
    pred_est_var = xds['{0}_comp'.format(pred_name)]
    pred_est_grad = xds['{0}_gradient_comp'.format(pred_name)]

    # use data inside timeframe
    dp_var = pred_est_var.values
    dp_grd = pred_est_grad.values
    shape_grid = dp_var[0].shape  # needed to handle data after PCs

    # unravel and join var and grad data 
    dp_ur = np.nan * np.ones(
        (dp_var.shape[0], 2*dp_var.shape[1]*dp_var.shape[2])
    )

    # we use .T to equal matlab
    for ti in range(dp_ur.shape[0]):
        dp_ur[ti,:] = np.concatenate(
            [np.ravel(dp_var[ti].T) , np.ravel(dp_grd[ti].T)]
        )

    # remove nans from predictor    
    data_pos = ~np.isnan(dp_ur[0,:])
    clean_row = dp_ur[0, data_pos]
    dp_ur_nonan = np.nan * np.ones(
        (dp_ur.shape[0], len(clean_row))
    )
    for ti in range(dp_ur.shape[0]):
        dp_ur_nonan[ti,:] = dp_ur[ti, data_pos]

    # standarize predictor
    pred_mean = np.mean(dp_ur_nonan, axis=0)
    pred_std = np.std(dp_ur_nonan, axis=0)
    pred_norm = (dp_ur_nonan[:,:] - pred_mean) / pred_std
    pred_norm[np.isnan(pred_norm)] = 0

    # principal components analysis
    ipca = PCA(n_components=min(pred_norm.shape[0], pred_norm.shape[1]))
    PCs = ipca.fit_transform(pred_norm)

    # return dataset
    return xr.Dataset(
        {
            'PCs': (('time', 'n_components'), PCs),
            'EOFs': (('n_components','n_features'), ipca.components_),
            'variance': (('n_components',), ipca.explained_variance_),

            'pred_mean': (('n_features',), pred_mean),
            'pred_std': (('n_features',), pred_std),

            'pred_lon': (('n_lon',), xds.longitude.values[:]),
            'pred_lat': (('n_lat',), xds.latitude.values[:]),
            'pred_time': (('time',), xds.time.values[:]),
            'pred_data_pos':(('n_points',), data_pos)
        },

        attrs = {
            'method': 'gradient + estela',
            'pred_name': pred_name,
        }
    )

